"""Implementation of the MediaRemoteTV Protocol used by ATV4 and later."""

import math
import logging
import asyncio
import datetime

from pyatv import exceptions, net
from pyatv.const import (
    Protocol, MediaType, DeviceState, RepeatState, ShuffleState)
from pyatv.cache import Cache
from pyatv.mrp import (messages, protobuf)
from pyatv.mrp.srp import SRPAuthHandler
from pyatv.mrp.connection import MrpConnection
from pyatv.mrp.protocol import MrpProtocol
from pyatv.mrp.protobuf import CommandInfo_pb2, SetStateMessage_pb2
from pyatv.mrp.player_state import PlayerStateManager
from pyatv.interface import (AppleTV, RemoteControl, Metadata,
                             Playing, PushUpdater, ArtworkInfo)


_LOGGER = logging.getLogger(__name__)

# Source: https://github.com/Daij-Djan/DDHidLib/blob/master/usb_hid_usages.txt
_KEY_LOOKUP = {
    # name: [usage_page, usage, button hold time (seconds)]
    'up': [1, 0x8C, 0],
    'down': [1, 0x8D, 0],
    'left': [1, 0x8B, 0],
    'right': [1, 0x8A, 0],
    'stop': [12, 0xB7, 0],
    'next': [12, 0xB5, 0],
    'previous': [12, 0xB6, 0],
    'select': [1, 0x89, 0],
    'menu': [1, 0x86, 0],
    'topmenu': [12, 0x60, 0],
    'home': [12, 0x40, 1],
    'suspend': [1, 0x82, 0],
    'wakeup': [1, 0x83, 0],
    'volume_up': [12, 0xE9, 0],
    'volume_down': [12, 0xEA, 0],

    # 'mic': [12, 0x04, 0],  # Siri
}


# pylint: disable=too-many-public-methods
class MrpRemoteControl(RemoteControl):
    """Implementation of API for controlling an Apple TV."""

    def __init__(self, loop, protocol):
        """Initialize a new MrpRemoteControl."""
        self.loop = loop
        self.protocol = protocol

    async def _press_key(self, key, hold=False):
        lookup = _KEY_LOOKUP.get(key, None)
        if lookup:
            await self.protocol.send(
                messages.send_hid_event(lookup[0], lookup[1], True))
            if hold:
                await asyncio.sleep(lookup[2])
            await self.protocol.send(
                messages.send_hid_event(lookup[0], lookup[1], False))
        else:
            raise Exception('unknown key: ' + key)

    def up(self):
        """Press key up."""
        return self._press_key('up')

    def down(self):
        """Press key down."""
        return self._press_key('down')

    def left(self):
        """Press key left."""
        return self._press_key('left')

    def right(self):
        """Press key right."""
        return self._press_key('right')

    def play(self):
        """Press key play."""
        return self.protocol.send(messages.command(CommandInfo_pb2.Play))

    def pause(self):
        """Press key play."""
        return self.protocol.send(messages.command(CommandInfo_pb2.Pause))

    def stop(self):
        """Press key stop."""
        return self.protocol.send(messages.command(CommandInfo_pb2.Stop))

    def next(self):
        """Press key next."""
        return self.protocol.send(messages.command(CommandInfo_pb2.NextTrack))

    def previous(self):
        """Press key previous."""
        return self.protocol.send(
            messages.command(CommandInfo_pb2.PreviousTrack))

    def select(self):
        """Press key select."""
        return self._press_key('select')

    def menu(self):
        """Press key menu."""
        return self._press_key('menu')

    def volume_up(self):
        """Press key volume up."""
        return self._press_key('volume_up')

    def volume_down(self):
        """Press key volume down."""
        return self._press_key('volume_down')

    def home(self):
        """Press key home."""
        return self._press_key('home')

    def home_hold(self):
        """Hold key home."""
        return self._press_key('home', hold=True)

    def top_menu(self):
        """Go to main menu (long press menu)."""
        return self._press_key('topmenu')

    def suspend(self):
        """Suspend the device."""
        return self._press_key('suspend')

    def wakeup(self):
        """Wake up the device."""
        return self._press_key('wakeup')

    def set_position(self, pos):
        """Seek in the current playing media."""
        return self.protocol.send(messages.seek_to_position(pos))

    def set_shuffle(self, shuffle_state):
        """Change shuffle mode to on or off."""
        return self.protocol.send(messages.shuffle(shuffle_state))

    def set_repeat(self, repeat_state):
        """Change repeat state."""
        return self.protocol.send(messages.repeat(repeat_state))


class MrpPlaying(Playing):
    """Implementation of API for retrieving what is playing."""

    def __init__(self, state):
        """Initialize a new MrpPlaying."""
        self._state = state

    @property
    def media_type(self):
        """Type of media is currently playing, e.g. video, music."""
        if self._state.metadata:
            media_type = self._state.metadata.mediaType
            cim = protobuf.ContentItemMetadata
            if media_type == cim.Audio:
                return MediaType.Music
            if media_type == cim.Video:
                return MediaType.Video

        return MediaType.Unknown

    @property
    def device_state(self):  # pylint: disable=too-many-return-statements
        """Device state, e.g. playing or paused."""
        state = self._state.playback_state
        ssm = SetStateMessage_pb2.SetStateMessage
        if state is None:
            return DeviceState.Idle
        if state == ssm.Playing:
            return DeviceState.Playing
        if state == ssm.Paused:
            return DeviceState.Paused
        if state == ssm.Stopped:
            return DeviceState.Stopped
        if state == ssm.Interrupted:
            return DeviceState.Loading
        if state == ssm.Seeking:
            return DeviceState.Seeking

        return DeviceState.Paused

    @property
    def title(self):
        """Title of the current media, e.g. movie or song name."""
        return self._state.metadata_field('title')

    @property
    def artist(self):
        """Artist of the currently playing song."""
        return self._state.metadata_field('trackArtistName')

    @property
    def album(self):
        """Album of the currently playing song."""
        return self._state.metadata_field('albumName')

    @property
    def genre(self):
        """Genre of the currently playing song."""
        return self._state.metadata_field('genre')

    @property
    def total_time(self):
        """Total play time in seconds."""
        duration = self._state.metadata_field('duration')
        if duration is None or math.isnan(duration):
            return None
        return int(duration)

    @property
    def position(self):
        """Position in the playing media (seconds)."""
        # If we don't have reference time, we can't do anything
        if not self._state.timestamp:
            return None

        elapsed_time = self._state.metadata_field('elapsedTime')
        now = datetime.datetime.now()
        diff = (now - self._state.timestamp).total_seconds()

        # If elapsed time is available, we make the assumption that
        # it is zero (playback started at reference time)
        elapsed_time = elapsed_time or 0
        if self.device_state == DeviceState.Playing:
            return int(elapsed_time + diff)
        return int(elapsed_time)

    def _get_command_info(self, command):
        for cmd in self._state.supported_commands:
            if cmd.command == command:
                return cmd
        return None

    @property
    def shuffle(self):
        """If shuffle is enabled or not."""
        info = self._get_command_info(CommandInfo_pb2.ChangeShuffleMode)
        if info is None:
            return ShuffleState.Off
        if info.shuffleMode == protobuf.CommandInfo.Off:
            return ShuffleState.Off
        if info.shuffleMode == protobuf.CommandInfo.Albums:
            return ShuffleState.Albums

        return ShuffleState.Songs

    @property
    def repeat(self):
        """Repeat mode."""
        info = self._get_command_info(CommandInfo_pb2.ChangeRepeatMode)
        if info is None:
            return RepeatState.Off
        if info.repeatMode == protobuf.CommandInfo.One:
            return RepeatState.Track

        return RepeatState.All

    @property
    def hash(self):
        """Create a unique hash for what is currently playing."""
        return self._state.item_identifier or super().hash


class MrpMetadata(Metadata):
    """Implementation of API for retrieving metadata."""

    def __init__(self, protocol, psm, identifier):
        """Initialize a new MrpPlaying."""
        super().__init__(identifier)
        self.protocol = protocol
        self.psm = psm
        self.artwork_cache = Cache(limit=4)

    async def artwork(self):
        """Return artwork for what is currently playing (or None)."""
        identifier = self.artwork_id
        if not identifier:
            _LOGGER.debug('No artwork available')
            return None

        if identifier in self.artwork_cache:
            _LOGGER.debug('Retrieved artwork %s from cache', identifier)
            return self.artwork_cache.get(identifier)

        artwork = await self._fetch_artwork()
        if artwork:
            self.artwork_cache.put(identifier, artwork)
            return artwork

        return None

    async def _fetch_artwork(self):
        playing = self.psm.playing
        resp = await self.psm.protocol.send_and_receive(
            messages.playback_queue_request(playing.location))
        if not resp.HasField('type'):
            return None

        item = resp.inner().playbackQueue.contentItems[playing.location]
        return ArtworkInfo(item.artworkData, playing.metadata.artworkMIMEType)

    @property
    def artwork_id(self):
        """Return a unique identifier for current artwork."""
        metadata = self.psm.playing.metadata
        if metadata and metadata.artworkAvailable:
            if metadata.HasField('artworkIdentifier'):
                return metadata.artworkIdentifier
            if metadata.HasField('contentIdentifier'):
                return metadata.contentIdentifier
        return None

    async def playing(self):
        """Return what is currently playing."""
        return MrpPlaying(self.psm.playing)


class MrpPushUpdater(PushUpdater):
    """Implementation of API for handling push update from an Apple TV."""

    def __init__(self, loop, metadata, psm):
        """Initialize a new MrpPushUpdater instance."""
        super().__init__()
        self.loop = loop
        self.metadata = metadata
        self.psm = psm
        self.listener = None

    @property
    def active(self):
        """Return if push updater has been started."""
        return self.psm.listener == self

    def start(self, initial_delay=0):
        """Wait for push updates from device.

        Will throw NoAsyncListenerError if no listener has been set.
        """
        if self.listener is None:
            raise exceptions.NoAsyncListenerError()
        if self.active:
            return

        self.psm.listener = self

    def stop(self):
        """No longer forward updates to listener."""
        self.psm.listener = None

    async def state_updated(self):
        """State was updated for active player."""
        try:
            playstatus = await self.metadata.playing()
            self.loop.call_soon(
                self.listener.playstatus_update, self, playstatus)
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.debug('Playstatus error occurred: %s', ex)
            self.loop.call_soon(self.listener.playstatus_error, self, ex)


class MrpAppleTV(AppleTV):
    """Implementation of API support for Apple TV."""

    # This is a container class so it's OK with many attributes
    # pylint: disable=too-many-instance-attributes
    def __init__(self, loop, session, config, airplay):
        """Initialize a new Apple TV."""
        super().__init__()

        self._session = session
        self._mrp_service = config.get_service(Protocol.MRP)

        self._connection = MrpConnection(
            config.address, self._mrp_service.port, loop, atv=self)
        self._srp = SRPAuthHandler()
        self._protocol = MrpProtocol(
            loop, self._connection, self._srp, self._mrp_service)
        self._psm = PlayerStateManager(self._protocol, loop)

        self._mrp_remote = MrpRemoteControl(loop, self._protocol)
        self._mrp_metadata = MrpMetadata(
            self._protocol, self._psm, config.identifier)
        self._mrp_push_updater = MrpPushUpdater(
            loop, self._mrp_metadata, self._psm)
        self._airplay = airplay

    async def connect(self):
        """Initiate connection to device.

        No need to call it yourself, it's done automatically.
        """
        await self._protocol.start()

    async def close(self):
        """Close connection and release allocated resources."""
        if net.is_custom_session(self._session):
            await self._session.close()
        self._protocol.stop()

    @property
    def service(self):
        """Return service used to connect to the Apple TV."""
        return self._mrp_service

    @property
    def remote_control(self):
        """Return API for controlling the Apple TV."""
        return self._mrp_remote

    @property
    def metadata(self):
        """Return API for retrieving metadata from Apple TV."""
        return self._mrp_metadata

    @property
    def push_updater(self):
        """Return API for handling push update from the Apple TV."""
        return self._mrp_push_updater

    @property
    def stream(self):
        """Return API for streaming media."""
        return self._airplay



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Optional as typing___Optional,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class RegisterHIDDeviceResultMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    errorCode = ... # type: int
    deviceIdentifier = ... # type: int

    def __init__(self,
        *,
        errorCode : typing___Optional[int] = None,
        deviceIdentifier : typing___Optional[int] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> RegisterHIDDeviceResultMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"deviceIdentifier",u"errorCode"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"deviceIdentifier",u"errorCode"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"deviceIdentifier",b"deviceIdentifier",u"errorCode",b"errorCode"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"deviceIdentifier",b"deviceIdentifier",u"errorCode",b"errorCode"]) -> None: ...

registerHIDDeviceResultMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



"""CLI application for discovering and controlling Apple TVs."""

import sys
import inspect
import logging
import binascii
import asyncio
import argparse
import traceback
from ipaddress import ip_address

from pyatv import (const, exceptions, interface, scan, connect, pair)
from pyatv.conf import (
    AppleTV, DmapService, MrpService, AirPlayService)
from pyatv.const import Protocol, ShuffleState, RepeatState
from pyatv.dmap import tag_definitions
from pyatv.dmap.parser import pprint
from pyatv.interface import retrieve_commands


def _print_commands(title, api):
    cmd_list = retrieve_commands(api)
    commands = ' - ' + '\n - '.join(
        map(lambda x: x[0] + ' - ' + x[1], sorted(cmd_list.items())))
    print('{} commands:\n{}\n'.format(title, commands))


async def _read_input(loop, prompt):
    sys.stdout.write(prompt)
    sys.stdout.flush()
    user_input = await loop.run_in_executor(None, sys.stdin.readline)
    return user_input.strip()


async def _scan_for_device(args, timeout, loop, protocol=None):
    options = {
        'timeout': timeout,
        'protocol': protocol
    }

    if not args.name:
        options['identifier'] = args.id
    if args.scan_hosts:
        options['hosts'] = args.scan_hosts

    atvs = await scan(loop, **options)

    if args.name:
        devices = [atv for atv in atvs if atv.name == args.name]
    else:
        devices = atvs

    if not atvs:
        logging.error('Could not find any Apple TV on current network')
        return None

    if len(devices) > 1:
        logging.error(
            'Found more than one Apple TV; specify one using --id')
        _print_found_apple_tvs(devices, sys.stderr)
        return None

    return devices[0]


class GlobalCommands:
    """Commands not bound to a specific device."""

    def __init__(self, args, loop):
        """Initialize a new instance of GlobalCommands."""
        self.args = args
        self.loop = loop

    async def commands(self):
        """Print a list with available commands."""
        _print_commands('Remote control', interface.RemoteControl)
        _print_commands('Metadata', interface.Metadata)
        _print_commands('Playing', interface.Playing)
        _print_commands('AirPlay', interface.Stream)
        _print_commands('Device', DeviceCommands)
        _print_commands('Global', self.__class__)

        return 0

    async def help(self):
        """Print help text for a command."""
        if len(self.args.command) != 2:
            print('Which command do you want help with?', file=sys.stderr)
            return 1

        iface = [interface.RemoteControl,
                 interface.Metadata,
                 interface.Playing,
                 interface.Stream,
                 self.__class__,
                 DeviceCommands]
        for cmd in iface:
            for key, value in cmd.__dict__.items():
                if key.startswith('_') or key != self.args.command[1]:
                    continue

                if inspect.isfunction(value):
                    signature = inspect.signature(value)
                else:
                    signature = ' (property)'

                print('COMMAND:\n>> {0}{1}\n\nHELP:\n{2}'.format(
                    key, signature, inspect.getdoc(value)))
        return 0

    async def scan(self):
        """Scan for Apple TVs on the network."""
        atvs = await scan(self.loop,
                          hosts=self.args.scan_hosts,
                          timeout=self.args.scan_timeout)
        _print_found_apple_tvs(atvs, sys.stdout)

        return 0

    async def pair(self):
        """Pair pyatv as a remote control with an Apple TV."""
        if self.args.protocol is None:
            logging.error('No protocol specified')
            return 1

        apple_tv = await _scan_for_device(
            self.args, self.args.scan_timeout, self.loop)
        if not apple_tv:
            return 2

        options = {}

        # Inject user provided credentials
        apple_tv.set_credentials(
            const.Protocol.AirPlay, self.args.airplay_credentials)
        apple_tv.set_credentials(
            const.Protocol.DMAP, self.args.dmap_credentials)
        apple_tv.set_credentials(
            const.Protocol.MRP, self.args.mrp_credentials)

        # Protocol specific options
        if self.args.protocol == const.Protocol.DMAP:
            options.update({
                'pairing_guid': self.args.pairing_guid,
                'remote_name': self.args.remote_name,
            })

        pairing = await pair(
            apple_tv, self.args.protocol, self.loop, **options)

        try:
            await self._perform_pairing(pairing)
        except Exception:  # pylint: disable=broad-except  # noqa
            logging.exception('Pairing failed')
            return 3
        finally:
            await pairing.close()

        return 0

    async def _perform_pairing(self, pairing):
        await pairing.begin()

        # Ask for PIN if present or just wait for pairing to end
        if pairing.device_provides_pin:
            pin = await _read_input(self.loop, 'Enter PIN on screen: ')
            pairing.pin(pin)
        else:
            pairing.pin(self.args.pin_code)

            if self.args.pin_code is None:
                print('Use any pin to pair with "{}"'
                      ' (press ENTER to stop)'.format(
                          self.args.remote_name))
            else:
                print('Use pin {} to pair with "{}"'
                      ' (press ENTER to stop)'.format(
                          self.args.pin_code, self.args.remote_name))

            await self.loop.run_in_executor(None, sys.stdin.readline)

        await pairing.finish()

        # Give some feedback to the user
        if pairing.has_paired:
            print('Pairing seems to have succeeded, yey!')
            print('You may now use these credentials: {0}'.format(
                pairing.service.credentials))
        else:
            print('Pairing failed!')


class DeviceCommands:
    """Additional commands available for a device.

    These commands are not part of the API but are provided by atvremote.
    """

    def __init__(self, atv, loop, args):
        """Initialize a new instance of DeviceCommands."""
        self.atv = atv
        self.loop = loop
        self.args = args

    async def cli(self):
        """Enter commands in a simple CLI."""
        print('Enter commands and press enter')
        print('Type help for help and exit to quit')

        while True:
            command = await _read_input(self.loop, 'pyatv> ')
            if command.lower() == 'exit':
                break

            if command == 'cli':
                print('Command not availble here')
                continue

            await _handle_device_command(
                self.args, command, self.atv, self.loop)

    async def artwork_save(self):
        """Download artwork and save it to artwork.png."""
        artwork = await self.atv.metadata.artwork()
        if artwork is not None:
            with open('artwork.png', 'wb') as file:
                file.write(artwork.bytes)
        else:
            print('No artwork is currently available.')
            return 1
        return 0

    async def push_updates(self):
        """Listen for push updates."""
        print('Press ENTER to stop')

        self.atv.push_updater.start()
        await self.loop.run_in_executor(None, sys.stdin.readline)
        self.atv.push_updater.stop()
        return 0


class PushListener:
    """Internal listener for push updates."""

    @staticmethod
    def playstatus_update(_, playstatus):
        """Print what is currently playing when it changes."""
        print(str(playstatus), flush=True)
        print(20*'-', flush=True)

    @staticmethod
    def playstatus_error(_, exception):
        """Inform about an error and restart push updates."""
        print("An error occurred (restarting): {0}".format(exception))


class DeviceListener(interface.DeviceListener):
    """Internal listener for generic device updates."""

    def connection_lost(self, exception):
        """Call when unexpectedly being disconnected from device."""
        print("Connection lost with error:", str(exception), file=sys.stderr)

    def connection_closed(self):
        """Call when connection was (intentionally) closed."""
        logging.debug("Connection was closed properly")


def _in_range(lower, upper, allow_none=False):
    def _checker(value):
        if allow_none and str(value).lower() == 'none':
            return None
        if int(value) >= lower and int(value) < upper:
            return int(value)
        raise argparse.ArgumentTypeError(
            'Must be greater >= {} and < {}'.format(lower, upper))
    return _checker


# pylint: disable=too-few-public-methods
class TransformProtocol(argparse.Action):
    """Transform protocol in string format to internal representation."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Match protocol string and save correct version."""
        if values == "mrp":
            setattr(namespace, self.dest, const.Protocol.MRP)
        elif values == "dmap":
            setattr(namespace, self.dest, const.Protocol.DMAP)
        elif values == 'airplay':
            setattr(namespace, self.dest, const.Protocol.AirPlay)
        else:
            raise argparse.ArgumentTypeError(
                'Valid protocols are: mrp, dmap, airplay')


# pylint: disable=too-few-public-methods
class TransformScanHosts(argparse.Action):
    """Transform scan hosts into array."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Split hosts and save as array."""
        ips = [ip_address(ip) for ip in values.split(',')]
        setattr(namespace, self.dest, ips)


async def cli_handler(loop):
    """Application starts here."""
    parser = argparse.ArgumentParser()

    parser.add_argument('command', nargs='+',
                        help='commands, help, ...')
    parser.add_argument('-i', '--id', help='device identifier',
                        dest='id', default=None)
    parser.add_argument('-n', '--name', help='apple tv name',
                        dest='name', default=None)
    parser.add_argument('--address', help='device ip address or hostname',
                        dest='address', default=None)
    parser.add_argument('--protocol', action=TransformProtocol,
                        help='protocol to use (values: dmap, mrp)',
                        dest='protocol', default=None)
    parser.add_argument('--port', help='port when connecting',
                        dest='port', type=_in_range(0, 65535),
                        default=0)
    parser.add_argument('-t', '--scan-timeout', help='timeout when scanning',
                        dest='scan_timeout', type=_in_range(1, 100),
                        metavar='TIMEOUT', default=3)
    parser.add_argument('-s', '--scan-hosts', help='scan specific hosts',
                        dest='scan_hosts', default=None,
                        action=TransformScanHosts)
    parser.add_argument('--version', action='version',
                        help='version of atvremote and pyatv',
                        version='%(prog)s {0}'.format(const.__version__))

    pairing = parser.add_argument_group('pairing')
    pairing.add_argument('--remote-name', help='remote pairing name',
                         dest='remote_name', default='pyatv')
    pairing.add_argument('-p', '--pin', help='pairing pin code',
                         dest='pin_code', metavar='PIN', default=1234,
                         type=_in_range(0, 9999, allow_none=True))
    pairing.add_argument('--pairing-guid',
                         help='pairing guid (16 chars hex)',
                         dest='pairing_guid', default=None)

    parser.add_argument('-m', '--manual', action='store_true',
                        help='use manual device details',
                        dest='manual', default=False)

    creds = parser.add_argument_group('credentials')
    creds.add_argument('--dmap-credentials', help='DMAP credentials to device',
                       dest='dmap_credentials', default=None)
    creds.add_argument('--mrp-credentials', help='MRP credentials to device',
                       dest='mrp_credentials', default=None)
    creds.add_argument('--airplay-credentials',
                       help='credentials for airplay',
                       dest='airplay_credentials', default=None)

    debug = parser.add_argument_group('debugging')
    debug.add_argument('-v', '--verbose', help='increase output verbosity',
                       action='store_true', dest='verbose')
    debug.add_argument('--debug', help='print debug information',
                       action='store_true', dest='debug')

    args = parser.parse_args()
    loglevel = logging.WARNING
    if args.verbose:
        loglevel = logging.INFO
    if args.debug:
        loglevel = logging.DEBUG

    logging.basicConfig(level=loglevel,
                        datefmt='%Y-%m-%d %H:%M:%S',
                        format='%(asctime)s %(levelname)s: %(message)s')
    logging.getLogger('requests').setLevel(logging.WARNING)

    cmds = retrieve_commands(GlobalCommands)

    if args.command[0] in cmds:
        glob_cmds = GlobalCommands(args, loop)
        return (await _exec_command(
            glob_cmds, args.command[0], print_result=False))
    if not args.manual:
        config = await _autodiscover_device(args, loop)
        if not config:
            return 1

        return await _handle_commands(args, config, loop)

    if args.port == 0 or args.address is None or args.protocol is None:
        logging.error(
            'You must specify address, port and protocol in manual mode')
        return 1

    config = _manual_device(args)
    return await _handle_commands(args, config, loop)


def _print_found_apple_tvs(atvs, outstream):
    print('Scan Results', file=outstream)
    print('=' * 40, file=outstream)
    for apple_tv in atvs:
        print('{0}\n'.format(apple_tv), file=outstream)


async def _autodiscover_device(args, loop):
    apple_tv = await _scan_for_device(
        args, args.scan_timeout, loop, protocol=args.protocol)
    if not apple_tv:
        return None

    def _set_credentials(protocol, field):
        service = apple_tv.get_service(protocol)
        if service:
            value = service.credentials or getattr(args, field)
            service.credentials = value

    _set_credentials(Protocol.DMAP, 'dmap_credentials')
    _set_credentials(Protocol.MRP, 'mrp_credentials')
    _set_credentials(Protocol.AirPlay, 'airplay_credentials')

    logging.info('Auto-discovered %s at %s', args.name, args.address)

    return apple_tv


def _manual_device(args):
    config = AppleTV(args.address, args.name)
    if args.dmap_credentials or args.protocol == const.Protocol.DMAP:
        config.add_service(DmapService(
            args.id, args.dmap_credentials, port=args.port))
    if args.mrp_credentials or args.protocol == const.Protocol.MRP:
        config.add_service(MrpService(
            args.id, args.port, credentials=args.mrp_credentials))
    if args.airplay_credentials:
        config.add_service(AirPlayService(
            args.id, credentials=args.airplay_credentials))
    return config


def _extract_command_with_args(cmd):
    """Parse input command with arguments.

    Parses the input command in such a way that the user may
    provide additional argument to the command. The format used is this:
      command=arg1,arg2,arg3,...
    all the additional arguments are passed as arguments to the target
    method.
    """
    def _typeparse(value):
        try:
            return int(value)
        except ValueError:
            return value

    def _parse_args(cmd, args):
        args = [_typeparse(x) for x in args]
        if cmd == 'set_shuffle':
            return [ShuffleState(args[0])]
        if cmd == 'set_repeat':
            return [RepeatState(args[0])]
        return args

    equal_sign = cmd.find('=')
    if equal_sign == -1:
        return cmd, []

    command = cmd[0:equal_sign]
    args = cmd[equal_sign+1:].split(',')
    return command, _parse_args(command, args)


async def _handle_commands(args, config, loop):
    atv = await connect(config, loop, protocol=args.protocol)
    atv.listener = DeviceListener()
    atv.push_updater.listener = PushListener()

    try:
        for cmd in args.command:
            ret = await _handle_device_command(args, cmd, atv, loop)
            if ret != 0:
                return ret
    finally:
        await atv.close()

    return 0


# pylint: disable=too-many-return-statements
async def _handle_device_command(args, cmd, atv, loop):
    device = retrieve_commands(DeviceCommands)
    ctrl = retrieve_commands(interface.RemoteControl)
    metadata = retrieve_commands(interface.Metadata)
    playing = retrieve_commands(interface.Playing)
    stream = retrieve_commands(interface.Stream)

    # Parse input command and argument from user
    cmd, cmd_args = _extract_command_with_args(cmd)
    if cmd in device:
        return (await _exec_command(
            DeviceCommands(atv, loop, args), cmd, False, *cmd_args))

    if cmd in ctrl:
        return (await _exec_command(
            atv.remote_control, cmd, True, *cmd_args))

    if cmd in metadata:
        return (await _exec_command(
            atv.metadata, cmd, True, *cmd_args))

    if cmd in playing:
        playing_resp = await atv.metadata.playing()
        return (await _exec_command(
            playing_resp, cmd, True, *cmd_args))

    if cmd in stream:
        return (await _exec_command(
            atv.stream, cmd, True, *cmd_args))

    logging.error('Unknown command: %s', cmd)
    return 1


async def _exec_command(obj, command, print_result, *args):
    try:
        # If the command to execute is a @property, the value returned by that
        # property will be stored in tmp. Otherwise it's a coroutine and we
        # have to yield for the result and wait until it is available.
        tmp = getattr(obj, command)
        if inspect.ismethod(tmp):
            value = await tmp(*args)
        else:
            value = tmp

        # Some commands might produce output themselves (especially non-API
        # commands), so don't print the return code they might give
        if print_result:
            _pretty_print(value)
            return 0
        return value
    except NotImplementedError:
        logging.exception("Command '%s' is not supported by device", command)
    except exceptions.AuthenticationError as ex:
        logging.exception('Authentication error: %s', str(ex))
    return 1


def _pretty_print(data):
    if data is None:
        return
    if isinstance(data, bytes):
        print(binascii.hexlify(data))
    elif isinstance(data, list):
        print(pprint(data, tag_definitions.lookup_tag))
    else:
        print(data)


async def appstart(loop):
    """Start the asyncio event loop and runs the application."""
    # Helper method so that the coroutine exits cleanly if an exception
    # happens (which would leave resources dangling)
    async def _run_application(loop):
        try:
            return await cli_handler(loop)

        except KeyboardInterrupt:
            pass  # User pressed Ctrl+C, just ignore it

        except SystemExit:
            pass  # sys.exit() was used - do nothing

        except Exception:  # pylint: disable=broad-except  # noqa
            traceback.print_exc(file=sys.stderr)
            sys.stderr.writelines(
                '\n>>> An error occurred, full stack trace above\n')

        return 1

    try:
        return await _run_application(loop)
    except KeyboardInterrupt:
        pass

    return 1


def main():
    """Application start here."""
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(appstart(loop))


if __name__ == '__main__':
    sys.exit(main())



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    Optional as typing___Optional,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class VirtualTouchDeviceDescriptor(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    absolute = ... # type: bool
    integratedDisplay = ... # type: bool
    screenSizeWidth = ... # type: float
    screenSizeHeight = ... # type: float

    def __init__(self,
        *,
        absolute : typing___Optional[bool] = None,
        integratedDisplay : typing___Optional[bool] = None,
        screenSizeWidth : typing___Optional[float] = None,
        screenSizeHeight : typing___Optional[float] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> VirtualTouchDeviceDescriptor: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"absolute",u"integratedDisplay",u"screenSizeHeight",u"screenSizeWidth"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"absolute",u"integratedDisplay",u"screenSizeHeight",u"screenSizeWidth"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"absolute",b"absolute",u"integratedDisplay",b"integratedDisplay",u"screenSizeHeight",b"screenSizeHeight",u"screenSizeWidth",b"screenSizeWidth"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"absolute",b"absolute",u"integratedDisplay",b"integratedDisplay",u"screenSizeHeight",b"screenSizeHeight",u"screenSizeWidth",b"screenSizeWidth"]) -> None: ...



"""Test suit for pairing process with Apple TV."""

import asynctest
import ipaddress

from pyatv import conf
from pyatv.dmap import (pairing, parser, tag_definitions)
from tests import zeroconf_stub, utils


REMOTE_NAME = 'pyatv remote'

# This is a valid config for default pairing guid
PIN_CODE = 1234
PAIRING_GUID = '0x0000000000000001'
PAIRING_CODE = '690E6FF61E0D7C747654A42AED17047D'

# This is valid for a some other (non-default) config
PIN_CODE2 = 5555
PAIRING_GUID2 = '0x1234ABCDE56789FF'
PAIRING_CODE2 = '58AD1D195B6DAA58AA2EA29DC25B81C3'

# Code is padded with zeros
PIN_CODE3 = 1
PAIRING_GUID3 = '0x7D1324235F535AE7'
PAIRING_CODE3 = 'A34C3361C7D57D61CA41F62A8042F069'

# Pairing guid is 8 bytes, which is 64 bits
RANDOM_128_BITS = 6558272190156386627
RANDOM_PAIRING_GUID = '0x5B03A9CF4A983143'
RANDOM_PAIRING_CODE = '7AF2D0B8629DE3C704D40A14C9E8CB93'


class PairingTest(asynctest.TestCase):

    async def setUp(self):
        self.service = conf.DmapService(None, None)
        self.config = conf.AppleTV('Apple TV', '127.0.0.1')
        self.config.add_service(self.service)
        self.zeroconf = zeroconf_stub.stub(pairing)
        self.pairing = None

        # TODO: currently stubs internal method, should provide stub
        # for netifaces later
        pairing._get_private_ip_addresses = \
            lambda: [ipaddress.ip_address('10.0.0.1')]

    async def tearDown(self):
        await self.pairing.finish()

    async def _start(self,
                     pin_code=PIN_CODE,
                     pairing_guid=PAIRING_GUID,
                     name=REMOTE_NAME):
        options = {'zeroconf': self.zeroconf}
        if pairing_guid:
            options['pairing_guid'] = pairing_guid
        if name:
            options['name'] = name

        self.pairing = pairing.DmapPairingHandler(
            self.config, None, self.loop, **options)
        await self.pairing.begin()
        self.pairing.pin(pin_code)

    async def test_zeroconf_service_published(self):
        await self._start()

        self.assertEqual(len(self.zeroconf.registered_services), 1,
                         msg='no zeroconf service registered')

        service = self.zeroconf.registered_services[0]
        self.assertEqual(service.properties[b'DvNm'], REMOTE_NAME,
                         msg='remote name does not match')

    async def test_succesful_pairing(self):
        await self._start()

        url = self._pairing_url(PAIRING_CODE)
        data, _ = await utils.simple_get(url)

        await self.pairing.finish()

        # Verify content returned in pairingresponse
        parsed = parser.parse(data, tag_definitions.lookup_tag)
        self.assertEqual(parser.first(parsed, 'cmpa', 'cmpg'), 1)
        self.assertEqual(parser.first(parsed, 'cmpa', 'cmnm'), REMOTE_NAME)
        self.assertEqual(parser.first(parsed, 'cmpa', 'cmty'), 'iPhone')

        self.assertEqual(self.service.credentials, PAIRING_GUID)

    async def test_successful_pairing_random_pairing_guid_generated(self):
        pairing.random.getrandbits = lambda x: RANDOM_128_BITS

        await self._start(pairing_guid=None)

        url = self._pairing_url(RANDOM_PAIRING_CODE)
        await utils.simple_get(url)

        await self.pairing.finish()

        self.assertEqual(self.service.credentials, RANDOM_PAIRING_GUID)

    async def test_succesful_pairing_with_any_pin(self):
        await self._start(pin_code=None)

        url = self._pairing_url('invalid_pairing_code')
        _, status = await utils.simple_get(url)

        self.assertEqual(status, 200)

    async def test_succesful_pairing_with_pin_leadering_zeros(self):
        await self._start(pin_code=PIN_CODE3, pairing_guid=PAIRING_GUID3)

        url = self._pairing_url(PAIRING_CODE3)
        _, status = await utils.simple_get(url)

        self.assertEqual(status, 200)

    async def test_pair_custom_pairing_guid(self):
        await self._start(pin_code=PIN_CODE2, pairing_guid=PAIRING_GUID2)

        url = self._pairing_url(PAIRING_CODE2)
        data, _ = await utils.simple_get(url)

        await self.pairing.finish()

        # Verify content returned in pairingresponse
        parsed = parser.parse(data, tag_definitions.lookup_tag)
        self.assertEqual(parser.first(parsed, 'cmpa', 'cmpg'),
                         int(PAIRING_GUID2, 16))

        self.assertEqual(self.service.credentials, PAIRING_GUID2)

    async def test_failed_pairing(self):
        await self._start()

        url = self._pairing_url('wrong')
        _, status = await utils.simple_get(url)

        self.assertEqual(status, 500)

    def _pairing_url(self, pairing_code):
        service = self.zeroconf.registered_services[0]
        server = 'http://127.0.0.1:{}'.format(service.port)
        return '{}/pairing?pairingcode={}&servicename=test'.format(
            server, pairing_code)



"""Functional tests for helper methods. Agnostic to protocol implementation."""

import asynctest

from asynctest.mock import patch

from pyatv import (conf, helpers)


class MockAppleTV:
    """Used to mock a device."""

    async def logout(self):
        """Fake logout method."""
        pass


class HelpersTest(asynctest.TestCase):

    def setUp(self):
        self.config = conf.AppleTV('address', 'name')
        self.mock_device = asynctest.mock.Mock(MockAppleTV())

    @patch('pyatv.scan', return_value=[])
    def test_auto_connect_with_no_device(self, scan_func):
        self.device_found = True

        async def found_handler():
            self.assertTrue(False, msg='should not be called')

        async def not_found_handler():
            self.device_found = False

        helpers.auto_connect(found_handler,
                             not_found=not_found_handler)

        self.assertFalse(self.device_found)

    @patch('pyatv.scan')
    @patch('pyatv.connect')
    def test_auto_connect_with_device(self, connect_func, scan_func):
        scan_func.return_value = [self.config]
        connect_func.return_value = self.mock_device

        self.found_device = None

        async def found_handler(atv):
            self.found_device = atv

        helpers.auto_connect(found_handler)

        self.assertEqual(self.found_device, self.mock_device)
        self.assertTrue(self.mock_device.logout.called)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    List as typing___List,
    Optional as typing___Optional,
    Text as typing___Text,
    Tuple as typing___Tuple,
    cast as typing___cast,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class ProtocolMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Type(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> ProtocolMessage.Type: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[ProtocolMessage.Type]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, ProtocolMessage.Type]]: ...
        SEND_COMMAND_MESSAGE = typing___cast(ProtocolMessage.Type, 1)
        SEND_COMMAND_RESULT_MESSAGE = typing___cast(ProtocolMessage.Type, 2)
        GET_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 3)
        SET_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 4)
        SET_ARTWORK_MESSAGE = typing___cast(ProtocolMessage.Type, 5)
        REGISTER_HID_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 6)
        REGISTER_HID_DEVICE_RESULT_MESSAGE = typing___cast(ProtocolMessage.Type, 7)
        SEND_HID_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 8)
        SEND_VIRTUAL_TOUCH_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 10)
        NOTIFICATION_MESSAGE = typing___cast(ProtocolMessage.Type, 11)
        CONTENT_ITEMS_CHANGED_NOTIFICATION_MESSAGE = typing___cast(ProtocolMessage.Type, 12)
        DEVICE_INFO_MESSAGE = typing___cast(ProtocolMessage.Type, 15)
        CLIENT_UPDATES_CONFIG_MESSAGE = typing___cast(ProtocolMessage.Type, 16)
        VOLUME_CONTROL_AVAILABILITY_MESSAGE = typing___cast(ProtocolMessage.Type, 17)
        GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 18)
        REGISTER_GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 19)
        REGISTER_GAME_CONTROLLER_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 20)
        UNREGISTER_GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 21)
        REGISTER_FOR_GAME_CONTROLLER_EVENTS_MESSAGE = typing___cast(ProtocolMessage.Type, 22)
        KEYBOARD_MESSAGE = typing___cast(ProtocolMessage.Type, 23)
        GET_KEYBOARD_SESSION_MESSAGE = typing___cast(ProtocolMessage.Type, 24)
        TEXT_INPUT_MESSAGE = typing___cast(ProtocolMessage.Type, 25)
        GET_VOICE_INPUT_DEVICES_MESSAGE = typing___cast(ProtocolMessage.Type, 26)
        GET_VOICE_INPUT_DEVICES_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 27)
        REGISTER_VOICE_INPUT_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 28)
        REGISTER_VOICE_INPUT_DEVICE_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 29)
        SET_RECORDING_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 30)
        SEND_VOICE_INPUT_MESSAGE = typing___cast(ProtocolMessage.Type, 31)
        PLAYBACK_QUEUE_REQUEST_MESSAGE = typing___cast(ProtocolMessage.Type, 32)
        TRANSACTION_MESSAGE = typing___cast(ProtocolMessage.Type, 33)
        CRYPTO_PAIRING_MESSAGE = typing___cast(ProtocolMessage.Type, 34)
        GAME_CONTROLLER_PROPERTIES_MESSAGE = typing___cast(ProtocolMessage.Type, 35)
        SET_READY_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 36)
        DEVICE_INFO_UPDATE_MESSAGE = typing___cast(ProtocolMessage.Type, 37)
        SET_CONNECTION_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 38)
        SET_HILITE_MODE_MESSAGE = typing___cast(ProtocolMessage.Type, 40)
        WAKE_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 41)
        SEND_PACKED_VIRTUAL_TOUCH_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 43)
        SET_NOW_PLAYING_CLIENT_MESSAGE = typing___cast(ProtocolMessage.Type, 46)
        UPDATE_CLIENT_MESSAGE = typing___cast(ProtocolMessage.Type, 55)
        UPDATE_CONTENT_ITEM_MESSAGE = typing___cast(ProtocolMessage.Type, 56)
    SEND_COMMAND_MESSAGE = typing___cast(ProtocolMessage.Type, 1)
    SEND_COMMAND_RESULT_MESSAGE = typing___cast(ProtocolMessage.Type, 2)
    GET_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 3)
    SET_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 4)
    SET_ARTWORK_MESSAGE = typing___cast(ProtocolMessage.Type, 5)
    REGISTER_HID_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 6)
    REGISTER_HID_DEVICE_RESULT_MESSAGE = typing___cast(ProtocolMessage.Type, 7)
    SEND_HID_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 8)
    SEND_VIRTUAL_TOUCH_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 10)
    NOTIFICATION_MESSAGE = typing___cast(ProtocolMessage.Type, 11)
    CONTENT_ITEMS_CHANGED_NOTIFICATION_MESSAGE = typing___cast(ProtocolMessage.Type, 12)
    DEVICE_INFO_MESSAGE = typing___cast(ProtocolMessage.Type, 15)
    CLIENT_UPDATES_CONFIG_MESSAGE = typing___cast(ProtocolMessage.Type, 16)
    VOLUME_CONTROL_AVAILABILITY_MESSAGE = typing___cast(ProtocolMessage.Type, 17)
    GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 18)
    REGISTER_GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 19)
    REGISTER_GAME_CONTROLLER_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 20)
    UNREGISTER_GAME_CONTROLLER_MESSAGE = typing___cast(ProtocolMessage.Type, 21)
    REGISTER_FOR_GAME_CONTROLLER_EVENTS_MESSAGE = typing___cast(ProtocolMessage.Type, 22)
    KEYBOARD_MESSAGE = typing___cast(ProtocolMessage.Type, 23)
    GET_KEYBOARD_SESSION_MESSAGE = typing___cast(ProtocolMessage.Type, 24)
    TEXT_INPUT_MESSAGE = typing___cast(ProtocolMessage.Type, 25)
    GET_VOICE_INPUT_DEVICES_MESSAGE = typing___cast(ProtocolMessage.Type, 26)
    GET_VOICE_INPUT_DEVICES_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 27)
    REGISTER_VOICE_INPUT_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 28)
    REGISTER_VOICE_INPUT_DEVICE_RESPONSE_MESSAGE = typing___cast(ProtocolMessage.Type, 29)
    SET_RECORDING_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 30)
    SEND_VOICE_INPUT_MESSAGE = typing___cast(ProtocolMessage.Type, 31)
    PLAYBACK_QUEUE_REQUEST_MESSAGE = typing___cast(ProtocolMessage.Type, 32)
    TRANSACTION_MESSAGE = typing___cast(ProtocolMessage.Type, 33)
    CRYPTO_PAIRING_MESSAGE = typing___cast(ProtocolMessage.Type, 34)
    GAME_CONTROLLER_PROPERTIES_MESSAGE = typing___cast(ProtocolMessage.Type, 35)
    SET_READY_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 36)
    DEVICE_INFO_UPDATE_MESSAGE = typing___cast(ProtocolMessage.Type, 37)
    SET_CONNECTION_STATE_MESSAGE = typing___cast(ProtocolMessage.Type, 38)
    SET_HILITE_MODE_MESSAGE = typing___cast(ProtocolMessage.Type, 40)
    WAKE_DEVICE_MESSAGE = typing___cast(ProtocolMessage.Type, 41)
    SEND_PACKED_VIRTUAL_TOUCH_EVENT_MESSAGE = typing___cast(ProtocolMessage.Type, 43)
    SET_NOW_PLAYING_CLIENT_MESSAGE = typing___cast(ProtocolMessage.Type, 46)
    UPDATE_CLIENT_MESSAGE = typing___cast(ProtocolMessage.Type, 55)
    UPDATE_CONTENT_ITEM_MESSAGE = typing___cast(ProtocolMessage.Type, 56)

    type = ... # type: ProtocolMessage.Type
    identifier = ... # type: typing___Text
    priority = ... # type: int
    timestamp = ... # type: int

    def __init__(self,
        *,
        type : typing___Optional[ProtocolMessage.Type] = None,
        identifier : typing___Optional[typing___Text] = None,
        priority : typing___Optional[int] = None,
        timestamp : typing___Optional[int] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> ProtocolMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"identifier",u"priority",u"timestamp",u"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"identifier",u"priority",u"timestamp",u"type"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"identifier",b"identifier",u"priority",b"priority",u"timestamp",b"timestamp",u"type",b"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"identifier",b"identifier",u"priority",b"priority",u"timestamp",b"timestamp",u"type",b"type"]) -> None: ...



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from pyatv.mrp.protobuf.DeviceInfoMessage_pb2 import (
    DeviceInfoMessage as pyatv___mrp___protobuf___DeviceInfoMessage_pb2___DeviceInfoMessage,
)

from typing import (
    List as typing___List,
    Optional as typing___Optional,
    Text as typing___Text,
    Tuple as typing___Tuple,
    cast as typing___cast,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class Origin(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class Type(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> Origin.Type: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[Origin.Type]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, Origin.Type]]: ...
        Unknown = typing___cast(Origin.Type, 0)
        Local = typing___cast(Origin.Type, 1)
        Custom = typing___cast(Origin.Type, 2)
    Unknown = typing___cast(Origin.Type, 0)
    Local = typing___cast(Origin.Type, 1)
    Custom = typing___cast(Origin.Type, 2)

    type = ... # type: Origin.Type
    displayName = ... # type: typing___Text
    identifier = ... # type: int

    @property
    def deviceInfo(self) -> pyatv___mrp___protobuf___DeviceInfoMessage_pb2___DeviceInfoMessage: ...

    def __init__(self,
        *,
        type : typing___Optional[Origin.Type] = None,
        displayName : typing___Optional[typing___Text] = None,
        identifier : typing___Optional[int] = None,
        deviceInfo : typing___Optional[pyatv___mrp___protobuf___DeviceInfoMessage_pb2___DeviceInfoMessage] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> Origin: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"deviceInfo",u"displayName",u"identifier",u"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"deviceInfo",u"displayName",u"identifier",u"type"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"deviceInfo",b"deviceInfo",u"displayName",b"displayName",u"identifier",b"identifier",u"type",b"type"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"deviceInfo",b"deviceInfo",u"displayName",b"displayName",u"identifier",b"identifier",u"type",b"type"]) -> None: ...



"""Various helper methods used by test cases."""

import time
import asyncio
import inspect

from datetime import datetime
from importlib import import_module

from aiohttp import ClientSession


def stub_sleep():
    """Stub asyncio.sleep and keep reference to real sleep.

    The tests should not sleep as that makes them slow, but the until function
    below must be able to do so (as it is polling). This method makes sure we
    still have a reference to original sleep in test environment.

    Calling this function stubs asyncio.sleep globally, so technically it only
    needs to be called once.
    """
    if not hasattr(asyncio, '_real_sleep'):
        # This is a special "hack" to schedule the sleep at the end of the
        # event queue in order to give other possibility to run.
        async def fake_sleep(time=None, loop=None):
            async def dummy():
                pass
            await asyncio.ensure_future(dummy())
        asyncio._real_sleep = asyncio.sleep
        asyncio.sleep = fake_sleep


async def simple_get(url):
    """Perform a GET-request to a specified URL."""
    session = ClientSession()
    response = await session.request('GET', url)
    if response.status < 200 or response.status >= 300:
        response.close()
        await session.close()
        return None, response.status

    data = await response.content.read()
    response.close()
    await session.close()
    return data, response.status


async def until(pred, timeout=5, **kwargs):
    """Wait until a predicate is fulfilled.

    Simple method of "waiting" for asynchronous code to finish.
    """
    deadline = time.time() + timeout
    while True:
        value = pred(**kwargs)
        if inspect.iscoroutinefunction(pred):
            value = await value

        if isinstance(value, tuple):
            cond, retvalue = value
            if cond:
                return retvalue
        elif value:
            return None

        if timeout is not None:
            if deadline - time.time() <= 0:
                raise asyncio.TimeoutError()

        # Use original method if stubbed
        if hasattr(asyncio, '_real_sleep'):
            await asyncio._real_sleep(0.5)
        else:
            await asyncio.sleep(0.5)


def faketime(module_name, *times):
    """Monkey patch datetime.now to return fake time."""
    class FakeDatetime:
        def __init__(self, times):
            self.times = times

        def __enter__(self):
            module = import_module(module_name)
            setattr(module.datetime, 'datetime', self)

        def __exit__(self, exc_type, exc_val, exc_tb):
            module = import_module(module_name)
            setattr(module.datetime, 'datetime', datetime)

        def now(self, tz=None):
            """Replace times from now to fake values."""
            if not self.time:
                return datetime.now(tz=tz)

            next_time = self.times[0]
            if len(self.times) > 1:
                self.times = self.times[1:]

            time = datetime.fromtimestamp(next_time)
            if tz:
                time = time.replace(tzinfo=tz)
            return time

        def __getattr__(self, attr):
            """Redirect non-stubbed functions to original module."""
            return getattr(datetime, attr)

    return FakeDatetime(list(times))



"""Common functional tests for all protocols."""

import logging
import asyncio

from aiohttp.test_utils import (AioHTTPTestCase, unittest_run_loop)

import pyatv
from pyatv import exceptions, interface
from pyatv.const import (
    Protocol, MediaType, DeviceState, RepeatState, ShuffleState)
from pyatv.conf import AppleTV, AirPlayService
from tests import zeroconf_stub
from tests.utils import stub_sleep, until, faketime


_LOGGER = logging.getLogger(__name__)

ARTWORK_BYTES = b'1234'
ARTWORK_BYTES2 = b'4321'
ARTWORK_MIMETYPE = 'image/png'
ARTWORK_ID = 'artwork_id1'
EXAMPLE_STREAM = 'http://stream'

HOMESHARING_SERVICE_1 = zeroconf_stub.homesharing_service(
    'AAAA', b'Apple TV 1', '10.0.0.1', b'aaaa')
HOMESHARING_SERVICE_2 = zeroconf_stub.homesharing_service(
    'BBBB', b'Apple TV 2', '10.0.0.2', b'bbbb')


async def poll(fn, **kwargs):
    result = await fn()
    conds = [getattr(result, f) == val for f, val in kwargs.items()]
    return all(conds), result


class DummyDeviceListener(interface.DeviceListener):

    def __init__(self):
        self.closed_sem = asyncio.Semaphore(0)
        self.lost_sem = asyncio.Semaphore(0)

    def connection_lost(self, exception):
        self.lost_sem.release()

    def connection_closed(self):
        self.closed_sem.release()


class CommonFunctionalTests(AioHTTPTestCase):

    def setUp(self):
        AioHTTPTestCase.setUp(self)
        stub_sleep()

    async def get_application(self, loop=None):
        raise NotImplementedError()

    async def playing(self, **kwargs):
        return await until(
            poll, fn=self.atv.metadata.playing, **kwargs)

    @unittest_run_loop
    async def test_connect_missing_device_id(self):
        conf = AppleTV('1.2.3.4', 'Apple TV')

        with self.assertRaises(exceptions.DeviceIdMissingError):
            await pyatv.connect(conf, self.loop)

    @unittest_run_loop
    async def test_connect_invalid_protocol(self):
        conf = AppleTV('1.2.3.4', 'Apple TV')
        conf.add_service(AirPlayService('airplay_id'))

        with self.assertRaises(exceptions.UnsupportedProtocolError):
            await pyatv.connect(
                conf, self.loop, protocol=Protocol.AirPlay)

    @unittest_run_loop
    async def test_pair_missing_service(self):
        conf = AppleTV('1.2.3.4', 'Apple TV')

        with self.assertRaises(exceptions.NoServiceError):
            await pyatv.pair(conf, Protocol.DMAP, self.loop)

    @unittest_run_loop
    async def test_play_url(self):
        self.usecase.airplay_playback_idle()
        self.usecase.airplay_playback_playing()
        self.usecase.airplay_playback_idle()

        await self.atv.stream.play_url(
            EXAMPLE_STREAM, port=self.server.port)

        self.assertEqual(self.fake_atv.last_airplay_url, EXAMPLE_STREAM)

    @unittest_run_loop
    async def test_play_url_not_authenticated_error(self):
        self.conf.get_service(Protocol.AirPlay).credentials = None
        self.usecase.airplay_require_authentication()

        with self.assertRaises(exceptions.AuthenticationError):
            await self.atv.stream.play_url(
                EXAMPLE_STREAM, port=self.server.port)

    @unittest_run_loop
    async def test_play_url_authenticated(self):
        self.usecase.airplay_require_authentication()
        self.usecase.airplay_playback_idle()
        self.usecase.airplay_playback_playing()
        self.usecase.airplay_playback_idle()

        await self.atv.stream.play_url(
            EXAMPLE_STREAM, port=self.server.port)

        self.assertEqual(self.fake_atv.last_airplay_url, EXAMPLE_STREAM)

    @unittest_run_loop
    async def test_button_up(self):
        await self.atv.remote_control.up()
        await until(lambda: self.fake_atv.last_button_pressed == 'up')

    @unittest_run_loop
    async def test_button_down(self):
        await self.atv.remote_control.down()
        await until(lambda: self.fake_atv.last_button_pressed == 'down')

    @unittest_run_loop
    async def test_button_left(self):
        await self.atv.remote_control.left()
        await until(lambda: self.fake_atv.last_button_pressed == 'left')

    @unittest_run_loop
    async def test_button_right(self):
        await self.atv.remote_control.right()
        await until(lambda: self.fake_atv.last_button_pressed == 'right')

    @unittest_run_loop
    async def test_button_select(self):
        await self.atv.remote_control.select()
        await until(lambda: self.fake_atv.last_button_pressed == 'select')

    @unittest_run_loop
    async def test_button_menu(self):
        await self.atv.remote_control.menu()
        await until(lambda: self.fake_atv.last_button_pressed == 'menu')

    @unittest_run_loop
    async def test_button_top_menu(self):
        await self.atv.remote_control.top_menu()
        await until(lambda: self.fake_atv.last_button_pressed == 'topmenu')

    @unittest_run_loop
    async def test_button_play(self):
        await self.atv.remote_control.play()
        await until(lambda: self.fake_atv.last_button_pressed == 'play')

    @unittest_run_loop
    async def test_button_pause(self):
        await self.atv.remote_control.pause()
        await until(lambda: self.fake_atv.last_button_pressed == 'pause')

    @unittest_run_loop
    async def test_button_stop(self):
        await self.atv.remote_control.stop()
        await until(lambda: self.fake_atv.last_button_pressed == 'stop')

    @unittest_run_loop
    async def test_button_next(self):
        await self.atv.remote_control.next()
        await until(lambda: self.fake_atv.last_button_pressed == 'nextitem')

    @unittest_run_loop
    async def test_button_previous(self):
        await self.atv.remote_control.previous()
        await until(lambda: self.fake_atv.last_button_pressed == 'previtem')

    def test_metadata_device_id(self):
        self.assertIn(self.atv.metadata.device_id, self.conf.all_identifiers)

    @unittest_run_loop
    async def test_close_connection(self):
        self.atv.listener = DummyDeviceListener()
        await self.atv.close()

        await asyncio.wait_for(
            self.atv.listener.closed_sem.acquire(), timeout=3.0)

    @unittest_run_loop
    async def test_metadata_video_paused(self):
        self.usecase.video_playing(paused=True, title='dummy',
                                   total_time=100, position=3)

        with faketime('pyatv', 0):
            playing = await self.playing(title='dummy')
            self.assertEqual(playing.media_type, MediaType.Video)
            self.assertEqual(playing.device_state, DeviceState.Paused)
            self.assertEqual(playing.title, 'dummy')
            self.assertEqual(playing.total_time, 100)
            self.assertEqual(playing.position, 3)

    @unittest_run_loop
    async def test_metadata_video_playing(self):
        self.usecase.video_playing(paused=False, title='video',
                                   total_time=40, position=10)

        with faketime('pyatv', 0):
            playing = await self.playing(title='video')
            self.assertEqual(playing.media_type, MediaType.Video)
            self.assertEqual(playing.device_state, DeviceState.Playing)
            self.assertEqual(playing.title, 'video')
            self.assertEqual(playing.total_time, 40)
            self.assertEqual(playing.position, 10)

    @unittest_run_loop
    async def test_metadata_music_paused(self):
        self.usecase.music_playing(paused=True, title='music',
                                   artist='artist', album='album',
                                   total_time=222, position=49,
                                   genre='genre')

        with faketime('pyatv', 0):
            playing = await self.playing(title='music')
            self.assertEqual(playing.media_type, MediaType.Music)
            self.assertEqual(playing.device_state, DeviceState.Paused)
            self.assertEqual(playing.title, 'music')
            self.assertEqual(playing.artist, 'artist')
            self.assertEqual(playing.album, 'album')
            self.assertEqual(playing.genre, 'genre')
            self.assertEqual(playing.total_time, 222)
            self.assertEqual(playing.position, 49)

    @unittest_run_loop
    async def test_metadata_music_playing(self):
        self.usecase.music_playing(paused=False, title='music',
                                   artist='test1', album='test2',
                                   total_time=2, position=1,
                                   genre='genre')

        with faketime('pyatv', 0):
            playing = await self.playing(title='music')
            self.assertEqual(playing.media_type, MediaType.Music)
            self.assertEqual(playing.device_state, DeviceState.Playing)
            self.assertEqual(playing.title, 'music')
            self.assertEqual(playing.artist, 'test1')
            self.assertEqual(playing.album, 'test2')
            self.assertEqual(playing.genre, 'genre')
            self.assertEqual(playing.total_time, 2)
            self.assertEqual(playing.position, 1)

    @unittest_run_loop
    async def test_seek_in_playing_media(self):
        self.usecase.video_playing(paused=False, title='dummy',
                                   total_time=40, position=10)

        with faketime('pyatv', 0):
            await self.atv.remote_control.set_position(30)
            playing = await self.playing(position=30)
            self.assertEqual(playing.position, 30)

    @unittest_run_loop
    async def test_metadata_artwork(self):
        self.usecase.example_video()
        self.usecase.change_artwork(ARTWORK_BYTES, ARTWORK_MIMETYPE)

        await self.playing(title='dummy')
        artwork = await self.atv.metadata.artwork()
        self.assertIsNotNone(artwork)
        self.assertEqual(artwork.bytes, ARTWORK_BYTES)
        self.assertEqual(artwork.mimetype, ARTWORK_MIMETYPE)

    @unittest_run_loop
    async def test_metadata_artwork_cache(self):
        self.usecase.example_video()
        self.usecase.change_artwork(
            ARTWORK_BYTES, ARTWORK_MIMETYPE, ARTWORK_ID)

        await self.playing(title='dummy')

        artwork = await self.atv.metadata.artwork()
        self.assertIsNotNone(artwork)
        self.assertEqual(artwork.bytes, ARTWORK_BYTES)

        artwork_id = self.atv.metadata.artwork_id

        # Change artwork data for same identifier (not really legal)
        self.usecase.change_artwork(
            ARTWORK_BYTES2, ARTWORK_MIMETYPE, ARTWORK_ID)

        # Expect previous data from cache
        artwork = await self.atv.metadata.artwork()
        self.assertIsNotNone(artwork)
        self.assertEqual(artwork.bytes, ARTWORK_BYTES)

        # Artwork identifier should be the same before as after
        self.assertEqual(self.atv.metadata.artwork_id, artwork_id)

    @unittest_run_loop
    async def test_push_updates(self):

        class PushListener:
            def __init__(self):
                self.playing = None

            def playstatus_update(self, updater, playstatus):
                _LOGGER.debug('Got playstatus: %s', playstatus)
                self.playing = playstatus

            @staticmethod
            def playstatus_error(updater, exception):
                pass

        # TODO: This test is a little weird as it leaks DMAP details
        # (revision). Should revise and make it better or use different tests
        # for each protocol.
        self.usecase.video_playing(paused=False, title='video1',
                                   total_time=40, position=10,
                                   revision=0)
        self.usecase.video_playing(paused=True, title='video2',
                                   total_time=30, position=20,
                                   revision=0)

        await self.playing()

        # Setup push updates which will instantly get the next one ("video2")
        listener = PushListener()
        self.atv.push_updater.listener = listener
        self.atv.push_updater.start()

        # Check that we got the right one
        await until(lambda: listener.playing is not None)
        self.assertEqual(listener.playing.title, 'video2')

    @unittest_run_loop
    async def test_push_updater_active(self):

        class DummyPushListener:

            @staticmethod
            def playstatus_update(updater, playstatus):
                pass

            @staticmethod
            def playstatus_error(updater, exception):
                pass

        self.usecase.video_playing(paused=False, title='video1',
                                   total_time=40, position=10,
                                   revision=0)

        self.assertFalse(self.atv.push_updater.active)

        self.atv.push_updater.listener = DummyPushListener()
        self.atv.push_updater.start()
        self.assertTrue(self.atv.push_updater.active)

        self.atv.push_updater.stop()
        self.assertFalse(self.atv.push_updater.active)

    @unittest_run_loop
    async def test_metadata_artwork_none_if_not_available(self):
        self.usecase.example_video()
        self.usecase.change_artwork(b'', None)

        await self.playing(title='dummy')
        artwork = await self.atv.metadata.artwork()
        self.assertIsNone(artwork)

    @unittest_run_loop
    async def test_metadata_none_type_when_not_playing(self):
        self.usecase.nothing_playing()

        playing = await self.playing()
        self.assertEqual(playing.media_type, MediaType.Unknown)
        self.assertEqual(playing.device_state, DeviceState.Idle)

    @unittest_run_loop
    async def test_repeat_state(self):
        for repeat in RepeatState:
            self.usecase.example_video(repeat=repeat)
            playing = await self.playing(repeat=repeat)
            self.assertEqual(playing.repeat, repeat)

    @unittest_run_loop
    async def test_set_repeat(self):
        self.usecase.video_playing(paused=False, title='video',
                                   total_time=40, position=10)
        for repeat in RepeatState:
            await self.atv.remote_control.set_repeat(repeat)
            playing = await self.playing(repeat=repeat)
            self.assertEqual(playing.repeat, repeat)

    @unittest_run_loop
    async def test_shuffle_state_common(self):
        for shuffle in [ShuffleState.Off, ShuffleState.Songs]:
            self.usecase.example_video(shuffle=shuffle)
            playing = await self.playing(shuffle=shuffle)
            self.assertEqual(playing.shuffle, shuffle)

    @unittest_run_loop
    async def test_set_shuffle_common(self):
        self.usecase.example_video()

        for shuffle in [ShuffleState.Off, ShuffleState.Songs]:
            await self.atv.remote_control.set_shuffle(shuffle)
            playing = await self.playing(shuffle=shuffle)
            self.assertEqual(playing.shuffle, shuffle)

    @unittest_run_loop
    async def test_metadata_loading(self):
        self.usecase.media_is_loading()

        playing = await self.playing(device_state=DeviceState.Loading)
        self.assertEqual(playing.device_state, DeviceState.Loading)



"""Functional tests for unicast DNS."""

import unittest
import asynctest

from pyatv import udns
from tests import fake_udns


MEDIAREMOTE_SERVICE = '_mediaremotetv._tcp.local'

TEST_SERVICES = {
  MEDIAREMOTE_SERVICE: fake_udns.FakeDnsService(
    name='Kitchen',
    port=1234,
    properties={
      'Name': 'Kitchen',
      'foo': '=bar',
    }),
}


def get_qtype(messages, qtype):
    for message in messages:
        if message.qtype == qtype:
            return message
    return None


class UdnsTest(unittest.TestCase):

    # This is a bastard test for the sake of testing labels (so I don't have to
    # implement it in the fake UDNS server).
    def test_qname_with_label(self):
        # This should resolve to "label.test" when reading from \x05
        message = b'aaaa' + b'\x04test\x00' + b'\x05label\xC0\x04\xAB\xCD'
        ptr = message[10:]
        ret, rest = udns.qname_decode(ptr, message)
        self.assertEqual(ret, 'label.test')
        self.assertEqual(rest, b'\xAB\xCD')


class UdnsFunctionalTest(asynctest.TestCase):

    async def setUp(self):
        self.server = fake_udns.FakeUdns(self.loop, TEST_SERVICES)
        await self.server.start()

    async def request(self, service_name):
        return await udns.request(
            self.loop, '127.0.0.1', [service_name], port=self.server.port,
            timeout=1), TEST_SERVICES.get(service_name)

    async def test_non_existing_service(self):
        resp, _ = await self.request('_missing')
        self.assertEqual(len(resp.questions), 1)
        self.assertEqual(len(resp.answers), 0)
        self.assertEqual(len(resp.resources), 0)

    async def test_service_has_expected_responses(self):
        resp, _ = await self.request(MEDIAREMOTE_SERVICE)
        self.assertEqual(len(resp.questions), 1)
        self.assertEqual(len(resp.answers), 1)
        self.assertEqual(len(resp.resources), 2)

    async def test_service_has_valid_question(self):
        resp, _ = await self.request(MEDIAREMOTE_SERVICE)

        question = resp.questions[0]
        self.assertEqual(question.qname, MEDIAREMOTE_SERVICE)
        self.assertEqual(question.qtype, udns.QTYPE_ANY)
        self.assertEqual(question.qclass, 0x8001)

    async def test_service_has_valid_answer(self):
        resp, data = await self.request(MEDIAREMOTE_SERVICE)

        answer = resp.answers[0]
        self.assertEqual(answer.qname, MEDIAREMOTE_SERVICE)
        self.assertEqual(answer.qtype, udns.QTYPE_PTR)
        self.assertEqual(answer.qclass, fake_udns.DEFAULT_QCLASS)
        self.assertEqual(answer.ttl, fake_udns.DEFAULT_TTL)
        self.assertEqual(answer.rd, data.name + '.' + MEDIAREMOTE_SERVICE)

    async def test_service_has_valid_srv_resource(self):
        resp, data = await self.request(MEDIAREMOTE_SERVICE)

        srv = get_qtype(resp.resources, udns.QTYPE_SRV)
        self.assertEqual(srv.qname, data.name + '.' + MEDIAREMOTE_SERVICE)
        self.assertEqual(srv.qtype, udns.QTYPE_SRV)
        self.assertEqual(srv.qclass, fake_udns.DEFAULT_QCLASS)
        self.assertEqual(srv.ttl, fake_udns.DEFAULT_TTL)

        rd = srv.rd
        self.assertEqual(rd['priority'], 0)
        self.assertEqual(rd['weight'], 0)
        self.assertEqual(rd['port'], data.port)
        self.assertEqual(rd['target'], data.name + '.local')

    async def test_service_has_valid_txt_resource(self):
        resp, data = await self.request(MEDIAREMOTE_SERVICE)

        srv = get_qtype(resp.resources, udns.QTYPE_TXT)
        self.assertEqual(srv.qname, data.name + '.' + MEDIAREMOTE_SERVICE)
        self.assertEqual(srv.qtype, udns.QTYPE_TXT)
        self.assertEqual(srv.qclass, fake_udns.DEFAULT_QCLASS)
        self.assertEqual(srv.ttl, fake_udns.DEFAULT_TTL)

        rd = srv.rd
        self.assertEqual(len(rd), len(data.properties))
        for k, v in data.properties.items():
            self.assertEqual(rd[k.encode('ascii')],
                             v.encode('ascii'))



"""Unit tests for pyatv.mrp.variant."""

import unittest

from pyatv.mrp.variant import (read_variant, write_variant)


class VariantTest(unittest.TestCase):

    def test_read_single_byte(self):
        self.assertEqual(read_variant(b'\x00')[0], 0x00)
        self.assertEqual(read_variant(b'\x35')[0], 0x35)

    def test_read_multiple_bytes(self):
        self.assertEqual(read_variant(b'\xb5\x44')[0], 8757)
        self.assertEqual(read_variant(b'\xc5\x92\x01')[0], 18757)

    def test_read_and_return_remaining_data(self):
        value, remaining = read_variant(b'\xb5\x44\xca\xfe')
        self.assertEqual(value, 8757)
        self.assertEqual(remaining, b'\xca\xfe')

    def test_read_invalid_variant(self):
        with self.assertRaises(Exception):
            read_variant(b'\x80')

    def test_write_single_byte(self):
        self.assertEqual(write_variant(0x00), b'\x00')
        self.assertEqual(write_variant(0x35), b'\x35')

    def test_write_multiple_bytes(self):
        self.assertEqual(write_variant(8757), b'\xb5\x44')
        self.assertEqual(write_variant(18757), b'\xc5\x92\x01')



"""Unit tests for pyatv.support."""

import asyncio
import unittest
import logging
from unittest.mock import MagicMock

import asynctest

from pyatv import exceptions
from pyatv.support import error_handler, log_binary


class TestException(Exception):
    pass


async def doraise(exception):
    raise exception()


class SupporErrorHandlerTest(asynctest.TestCase):

    async def test_return_value(self):
        async def _returns():
            return 123

        self.assertEqual(
            await error_handler(_returns, TestException), 123)

    async def test_oserror(self):
        with self.assertRaises(exceptions.ConnectionFailedError):
            await error_handler(doraise, TestException, OSError)

    async def test_timeout(self):
        with self.assertRaises(exceptions.ConnectionFailedError):
            await error_handler(
                doraise, TestException, asyncio.TimeoutError)

    async def test_backoff(self):
        with self.assertRaises(exceptions.BackOffError):
            await error_handler(
                doraise, TestException, exceptions.BackOffError)

    async def test_no_credentials(self):
        with self.assertRaises(exceptions.NoCredentialsError):
            await error_handler(
                doraise, TestException, exceptions.NoCredentialsError)

    async def test_other_exception(self):
        with self.assertRaises(TestException):
            await error_handler(doraise, TestException, Exception)


class SupporLogtTest(unittest.TestCase):

    def setUp(self):
        self.mock_logger = MagicMock()
        self.mock_logger.return_value = True

    def test_no_log_if_not_debug(self):
        self.mock_logger.isEnabledFor.return_value = False
        log_binary(self.mock_logger, 'test')
        self.mock_logger.isEnabledFor.assert_called_with(logging.DEBUG)

    def test_log_no_args_if_enabled(self):
        log_binary(self.mock_logger, 'testing')
        self.assertEqual(self._debug_string(), 'testing ()')

    def test_log_single_arg_if_enabled(self):
        log_binary(self.mock_logger, 'abc', test=b'\x01\x02')
        self.assertEqual(self._debug_string(), 'abc (test=0102)')

    def test_log_multiple_args_if_enabled(self):
        log_binary(self.mock_logger, 'k', test=b'\x01\x02', dummy=b'\xfe')
        self.assertEqual(self._debug_string(), 'k (dummy=fe, test=0102)')

    # This method generates the formatted debug string
    def _debug_string(self):
        call_args = self.mock_logger.debug.call_args[0]
        formatting = call_args[0]
        args = call_args[1:]
        return formatting % args



"""Simplified extension handling for protobuf messages.

THIS CODE IS AUTO-GENERATED - DO NOT EDIT!!!
"""

from pyatv.mrp.protobuf.ProtocolMessage_pb2 import ProtocolMessage


from pyatv.mrp.protobuf import ClientUpdatesConfigMessage_pb2
from pyatv.mrp.protobuf import CryptoPairingMessage_pb2
from pyatv.mrp.protobuf import DeviceInfoMessage_pb2
from pyatv.mrp.protobuf import GetKeyboardSessionMessage_pb2
from pyatv.mrp.protobuf import KeyboardMessage_pb2
from pyatv.mrp.protobuf import NotificationMessage_pb2
from pyatv.mrp.protobuf import PlaybackQueueRequestMessage_pb2
from pyatv.mrp.protobuf import RegisterForGameControllerEventsMessage_pb2
from pyatv.mrp.protobuf import RegisterHIDDeviceMessage_pb2
from pyatv.mrp.protobuf import RegisterHIDDeviceResultMessage_pb2
from pyatv.mrp.protobuf import RegisterVoiceInputDeviceMessage_pb2
from pyatv.mrp.protobuf import RegisterVoiceInputDeviceResponseMessage_pb2
from pyatv.mrp.protobuf import SendCommandMessage_pb2
from pyatv.mrp.protobuf import SendCommandResultMessage_pb2
from pyatv.mrp.protobuf import SendHIDEventMessage_pb2
from pyatv.mrp.protobuf import SendPackedVirtualTouchEventMessage_pb2
from pyatv.mrp.protobuf import SetArtworkMessage_pb2
from pyatv.mrp.protobuf import SetConnectionStateMessage_pb2
from pyatv.mrp.protobuf import SetHiliteModeMessage_pb2
from pyatv.mrp.protobuf import SetNowPlayingClientMessage_pb2
from pyatv.mrp.protobuf import SetStateMessage_pb2
from pyatv.mrp.protobuf import TextInputMessage_pb2
from pyatv.mrp.protobuf import TransactionMessage_pb2
from pyatv.mrp.protobuf import UpdateClientMessage_pb2
from pyatv.mrp.protobuf import UpdateContentItemMessage_pb2
from pyatv.mrp.protobuf import VolumeControlAvailabilityMessage_pb2
from pyatv.mrp.protobuf import WakeDeviceMessage_pb2


from pyatv.mrp.protobuf.AudioFormatSettingsMessage_pb2 import AudioFormatSettings
from pyatv.mrp.protobuf.ClientUpdatesConfigMessage_pb2 import ClientUpdatesConfigMessage
from pyatv.mrp.protobuf.CommandInfo_pb2 import CommandInfo
from pyatv.mrp.protobuf.CommandOptions_pb2 import CommandOptions
from pyatv.mrp.protobuf.ContentItemMetadata_pb2 import ContentItemMetadata
from pyatv.mrp.protobuf.ContentItem_pb2 import ContentItem
from pyatv.mrp.protobuf.CryptoPairingMessage_pb2 import CryptoPairingMessage
from pyatv.mrp.protobuf.DeviceInfoMessage_pb2 import DeviceInfoMessage
from pyatv.mrp.protobuf.GetKeyboardSessionMessage_pb2 import GetKeyboardSessionMessage
from pyatv.mrp.protobuf.KeyboardMessage_pb2 import KeyboardMessage
from pyatv.mrp.protobuf.LanguageOption_pb2 import LanguageOption
from pyatv.mrp.protobuf.NotificationMessage_pb2 import NotificationMessage
from pyatv.mrp.protobuf.NowPlayingClient_pb2 import NowPlayingClient
from pyatv.mrp.protobuf.NowPlayingInfo_pb2 import NowPlayingInfo
from pyatv.mrp.protobuf.NowPlayingPlayer_pb2 import NowPlayingPlayer
from pyatv.mrp.protobuf.Origin_pb2 import Origin
from pyatv.mrp.protobuf.PlaybackQueueCapabilities_pb2 import PlaybackQueueCapabilities
from pyatv.mrp.protobuf.PlaybackQueueContext_pb2 import PlaybackQueueContext
from pyatv.mrp.protobuf.PlaybackQueueRequestMessage_pb2 import PlaybackQueueRequestMessage
from pyatv.mrp.protobuf.PlaybackQueue_pb2 import PlaybackQueue
from pyatv.mrp.protobuf.PlayerPath_pb2 import PlayerPath
from pyatv.mrp.protobuf.RegisterForGameControllerEventsMessage_pb2 import RegisterForGameControllerEventsMessage
from pyatv.mrp.protobuf.RegisterHIDDeviceMessage_pb2 import RegisterHIDDeviceMessage
from pyatv.mrp.protobuf.RegisterHIDDeviceResultMessage_pb2 import RegisterHIDDeviceResultMessage
from pyatv.mrp.protobuf.RegisterVoiceInputDeviceMessage_pb2 import RegisterVoiceInputDeviceMessage
from pyatv.mrp.protobuf.RegisterVoiceInputDeviceResponseMessage_pb2 import RegisterVoiceInputDeviceResponseMessage
from pyatv.mrp.protobuf.SendCommandMessage_pb2 import SendCommandMessage
from pyatv.mrp.protobuf.SendCommandResultMessage_pb2 import SendCommandResultMessage
from pyatv.mrp.protobuf.SendHIDEventMessage_pb2 import SendHIDEventMessage
from pyatv.mrp.protobuf.SendPackedVirtualTouchEventMessage_pb2 import SendPackedVirtualTouchEventMessage
from pyatv.mrp.protobuf.SetArtworkMessage_pb2 import SetArtworkMessage
from pyatv.mrp.protobuf.SetConnectionStateMessage_pb2 import SetConnectionStateMessage
from pyatv.mrp.protobuf.SetHiliteModeMessage_pb2 import SetHiliteModeMessage
from pyatv.mrp.protobuf.SetNowPlayingClientMessage_pb2 import SetNowPlayingClientMessage
from pyatv.mrp.protobuf.SetStateMessage_pb2 import SetStateMessage
from pyatv.mrp.protobuf.SupportedCommands_pb2 import SupportedCommands
from pyatv.mrp.protobuf.TextEditingAttributesMessage_pb2 import TextEditingAttributes
from pyatv.mrp.protobuf.TextInputMessage_pb2 import TextInputMessage
from pyatv.mrp.protobuf.TextInputTraitsMessage_pb2 import TextInputTraits
from pyatv.mrp.protobuf.TransactionKey_pb2 import TransactionKey
from pyatv.mrp.protobuf.TransactionMessage_pb2 import TransactionMessage
from pyatv.mrp.protobuf.TransactionPacket_pb2 import TransactionPacket
from pyatv.mrp.protobuf.TransactionPackets_pb2 import TransactionPackets
from pyatv.mrp.protobuf.UpdateClientMessage_pb2 import UpdateClientMessage
from pyatv.mrp.protobuf.UpdateContentItemMessage_pb2 import UpdateContentItemMessage
from pyatv.mrp.protobuf.VirtualTouchDeviceDescriptorMessage_pb2 import VirtualTouchDeviceDescriptor
from pyatv.mrp.protobuf.VoiceInputDeviceDescriptorMessage_pb2 import VoiceInputDeviceDescriptor
from pyatv.mrp.protobuf.VolumeControlAvailabilityMessage_pb2 import VolumeControlAvailabilityMessage
from pyatv.mrp.protobuf.WakeDeviceMessage_pb2 import WakeDeviceMessage


_EXTENSION_LOOKUP = {
    ProtocolMessage.CLIENT_UPDATES_CONFIG_MESSAGE: ClientUpdatesConfigMessage_pb2.clientUpdatesConfigMessage,
    ProtocolMessage.CRYPTO_PAIRING_MESSAGE: CryptoPairingMessage_pb2.cryptoPairingMessage,
    ProtocolMessage.DEVICE_INFO_MESSAGE: DeviceInfoMessage_pb2.deviceInfoMessage,
    ProtocolMessage.DEVICE_INFO_UPDATE_MESSAGE: DeviceInfoMessage_pb2.deviceInfoMessage,
    ProtocolMessage.GET_KEYBOARD_SESSION_MESSAGE: GetKeyboardSessionMessage_pb2.getKeyboardSessionMessage,
    ProtocolMessage.KEYBOARD_MESSAGE: KeyboardMessage_pb2.keyboardMessage,
    ProtocolMessage.NOTIFICATION_MESSAGE: NotificationMessage_pb2.notificationMessage,
    ProtocolMessage.PLAYBACK_QUEUE_REQUEST_MESSAGE: PlaybackQueueRequestMessage_pb2.playbackQueueRequestMessage,
    ProtocolMessage.REGISTER_FOR_GAME_CONTROLLER_EVENTS_MESSAGE: RegisterForGameControllerEventsMessage_pb2.registerForGameControllerEventsMessage,
    ProtocolMessage.REGISTER_HID_DEVICE_MESSAGE: RegisterHIDDeviceMessage_pb2.registerHIDDeviceMessage,
    ProtocolMessage.REGISTER_HID_DEVICE_RESULT_MESSAGE: RegisterHIDDeviceResultMessage_pb2.registerHIDDeviceResultMessage,
    ProtocolMessage.REGISTER_VOICE_INPUT_DEVICE_MESSAGE: RegisterVoiceInputDeviceMessage_pb2.registerVoiceInputDeviceMessage,
    ProtocolMessage.REGISTER_VOICE_INPUT_DEVICE_RESPONSE_MESSAGE: RegisterVoiceInputDeviceResponseMessage_pb2.registerVoiceInputDeviceResponseMessage,
    ProtocolMessage.SEND_COMMAND_MESSAGE: SendCommandMessage_pb2.sendCommandMessage,
    ProtocolMessage.SEND_COMMAND_RESULT_MESSAGE: SendCommandResultMessage_pb2.sendCommandResultMessage,
    ProtocolMessage.SEND_HID_EVENT_MESSAGE: SendHIDEventMessage_pb2.sendHIDEventMessage,
    ProtocolMessage.SEND_PACKED_VIRTUAL_TOUCH_EVENT_MESSAGE: SendPackedVirtualTouchEventMessage_pb2.sendPackedVirtualTouchEventMessage,
    ProtocolMessage.SET_ARTWORK_MESSAGE: SetArtworkMessage_pb2.setArtworkMessage,
    ProtocolMessage.SET_CONNECTION_STATE_MESSAGE: SetConnectionStateMessage_pb2.setConnectionStateMessage,
    ProtocolMessage.SET_HILITE_MODE_MESSAGE: SetHiliteModeMessage_pb2.setHiliteModeMessage,
    ProtocolMessage.SET_NOW_PLAYING_CLIENT_MESSAGE: SetNowPlayingClientMessage_pb2.setNowPlayingClientMessage,
    ProtocolMessage.SET_STATE_MESSAGE: SetStateMessage_pb2.setStateMessage,
    ProtocolMessage.TEXT_INPUT_MESSAGE: TextInputMessage_pb2.textInputMessage,
    ProtocolMessage.TRANSACTION_MESSAGE: TransactionMessage_pb2.transactionMessage,
    ProtocolMessage.UPDATE_CLIENT_MESSAGE: UpdateClientMessage_pb2.updateClientMessage,
    ProtocolMessage.UPDATE_CONTENT_ITEM_MESSAGE: UpdateContentItemMessage_pb2.updateContentItemMessage,
    ProtocolMessage.VOLUME_CONTROL_AVAILABILITY_MESSAGE: VolumeControlAvailabilityMessage_pb2.volumeControlAvailabilityMessage,
    ProtocolMessage.WAKE_DEVICE_MESSAGE: WakeDeviceMessage_pb2.wakeDeviceMessage,
}


CLIENT_UPDATES_CONFIG_MESSAGE = ProtocolMessage.CLIENT_UPDATES_CONFIG_MESSAGE
CRYPTO_PAIRING_MESSAGE = ProtocolMessage.CRYPTO_PAIRING_MESSAGE
DEVICE_INFO_MESSAGE = ProtocolMessage.DEVICE_INFO_MESSAGE
GET_KEYBOARD_SESSION_MESSAGE = ProtocolMessage.GET_KEYBOARD_SESSION_MESSAGE
KEYBOARD_MESSAGE = ProtocolMessage.KEYBOARD_MESSAGE
NOTIFICATION_MESSAGE = ProtocolMessage.NOTIFICATION_MESSAGE
PLAYBACK_QUEUE_REQUEST_MESSAGE = ProtocolMessage.PLAYBACK_QUEUE_REQUEST_MESSAGE
REGISTER_FOR_GAME_CONTROLLER_EVENTS_MESSAGE = ProtocolMessage.REGISTER_FOR_GAME_CONTROLLER_EVENTS_MESSAGE
REGISTER_HID_DEVICE_MESSAGE = ProtocolMessage.REGISTER_HID_DEVICE_MESSAGE
REGISTER_HID_DEVICE_RESULT_MESSAGE = ProtocolMessage.REGISTER_HID_DEVICE_RESULT_MESSAGE
REGISTER_VOICE_INPUT_DEVICE_MESSAGE = ProtocolMessage.REGISTER_VOICE_INPUT_DEVICE_MESSAGE
REGISTER_VOICE_INPUT_DEVICE_RESPONSE_MESSAGE = ProtocolMessage.REGISTER_VOICE_INPUT_DEVICE_RESPONSE_MESSAGE
SEND_COMMAND_MESSAGE = ProtocolMessage.SEND_COMMAND_MESSAGE
SEND_COMMAND_RESULT_MESSAGE = ProtocolMessage.SEND_COMMAND_RESULT_MESSAGE
SEND_HID_EVENT_MESSAGE = ProtocolMessage.SEND_HID_EVENT_MESSAGE
SEND_PACKED_VIRTUAL_TOUCH_EVENT_MESSAGE = ProtocolMessage.SEND_PACKED_VIRTUAL_TOUCH_EVENT_MESSAGE
SET_ARTWORK_MESSAGE = ProtocolMessage.SET_ARTWORK_MESSAGE
SET_CONNECTION_STATE_MESSAGE = ProtocolMessage.SET_CONNECTION_STATE_MESSAGE
SET_HILITE_MODE_MESSAGE = ProtocolMessage.SET_HILITE_MODE_MESSAGE
SET_NOW_PLAYING_CLIENT_MESSAGE = ProtocolMessage.SET_NOW_PLAYING_CLIENT_MESSAGE
SET_STATE_MESSAGE = ProtocolMessage.SET_STATE_MESSAGE
TEXT_INPUT_MESSAGE = ProtocolMessage.TEXT_INPUT_MESSAGE
TRANSACTION_MESSAGE = ProtocolMessage.TRANSACTION_MESSAGE
UPDATE_CLIENT_MESSAGE = ProtocolMessage.UPDATE_CLIENT_MESSAGE
UPDATE_CONTENT_ITEM_MESSAGE = ProtocolMessage.UPDATE_CONTENT_ITEM_MESSAGE
VOLUME_CONTROL_AVAILABILITY_MESSAGE = ProtocolMessage.VOLUME_CONTROL_AVAILABILITY_MESSAGE
WAKE_DEVICE_MESSAGE = ProtocolMessage.WAKE_DEVICE_MESSAGE


def _inner_message(self):
    extension = _EXTENSION_LOOKUP.get(self.type, None)
    if extension:
        return self.Extensions[extension]

    raise Exception('unknown type: ' + str(self.type))


ProtocolMessage.inner = _inner_message  # type: ignore




"""Functional authentication tests with fake MRP Apple TV."""

from aiohttp.test_utils import (AioHTTPTestCase, unittest_run_loop)

import pyatv
from pyatv import exceptions
from pyatv.const import Protocol
from pyatv.conf import (MrpService, AppleTV)
from tests.mrp.fake_mrp_atv import (FakeAppleTV, AppleTVUseCases)
from tests.mrp.mrp_server_auth import (
    PIN_CODE, CLIENT_IDENTIFIER, CLIENT_CREDENTIALS)


class MrpAuthFunctionalTest(AioHTTPTestCase):

    def setUp(self):
        AioHTTPTestCase.setUp(self)

        self.service = MrpService(
            CLIENT_IDENTIFIER, self.fake_atv.port)
        self.conf = AppleTV('127.0.0.1', 'Apple TV')
        self.conf.add_service(self.service)

    async def tearDownAsync(self):
        await self.handle.close()
        await super().tearDownAsync()

    async def get_application(self, loop=None):
        self.fake_atv = FakeAppleTV(self, self.loop)
        self.usecase = AppleTVUseCases(self.fake_atv)
        return self.fake_atv.app

    @unittest_run_loop
    async def test_pairing_with_device(self):
        self.handle = await pyatv.pair(
            self.conf, Protocol.MRP, self.loop)

        self.assertTrue(self.handle.device_provides_pin)

        await self.handle.begin()
        self.handle.pin(PIN_CODE)

        await self.handle.finish()

        self.assertTrue(self.handle.has_paired)
        self.assertTrue(self.fake_atv.has_paired)

    @unittest_run_loop
    async def test_pairing_with_bad_pin(self):
        self.handle = await pyatv.pair(
            self.conf, Protocol.MRP, self.loop)

        self.assertTrue(self.handle.device_provides_pin)

        await self.handle.begin()
        self.handle.pin(PIN_CODE + 1)

        with self.assertRaises(exceptions.PairingError):
            await self.handle.finish()

        self.assertFalse(self.handle.has_paired)
        self.assertFalse(self.fake_atv.has_paired)

    @unittest_run_loop
    async def test_pairing_authentication(self):
        self.service.credentials = CLIENT_CREDENTIALS

        self.handle = await pyatv.connect(self.conf, self.loop)

        self.assertTrue(self.fake_atv.has_authenticated)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from pyatv.mrp.protobuf.TextEditingAttributesMessage_pb2 import (
    TextEditingAttributes as pyatv___mrp___protobuf___TextEditingAttributesMessage_pb2___TextEditingAttributes,
)

from typing import (
    Optional as typing___Optional,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class KeyboardMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    state = ... # type: int
    encryptedTextCyphertext = ... # type: bytes

    @property
    def attributes(self) -> pyatv___mrp___protobuf___TextEditingAttributesMessage_pb2___TextEditingAttributes: ...

    def __init__(self,
        *,
        state : typing___Optional[int] = None,
        attributes : typing___Optional[pyatv___mrp___protobuf___TextEditingAttributesMessage_pb2___TextEditingAttributes] = None,
        encryptedTextCyphertext : typing___Optional[bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> KeyboardMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"attributes",u"encryptedTextCyphertext",u"state"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"attributes",u"encryptedTextCyphertext",u"state"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"attributes",b"attributes",u"encryptedTextCyphertext",b"encryptedTextCyphertext",u"state",b"state"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"attributes",b"attributes",u"encryptedTextCyphertext",b"encryptedTextCyphertext",u"state",b"state"]) -> None: ...

keyboardMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    EnumDescriptor as google___protobuf___descriptor___EnumDescriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)

from typing import (
    List as typing___List,
    Optional as typing___Optional,
    Text as typing___Text,
    Tuple as typing___Tuple,
    cast as typing___cast,
)

from typing_extensions import (
    Literal as typing_extensions___Literal,
)


class TextInputMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...
    class ActionType(int):
        DESCRIPTOR: google___protobuf___descriptor___EnumDescriptor = ...
        @classmethod
        def Name(cls, number: int) -> str: ...
        @classmethod
        def Value(cls, name: str) -> TextInputMessage.ActionType: ...
        @classmethod
        def keys(cls) -> typing___List[str]: ...
        @classmethod
        def values(cls) -> typing___List[TextInputMessage.ActionType]: ...
        @classmethod
        def items(cls) -> typing___List[typing___Tuple[str, TextInputMessage.ActionType]]: ...
        Unknown = typing___cast(TextInputMessage.ActionType, 0)
        Insert = typing___cast(TextInputMessage.ActionType, 1)
        Set = typing___cast(TextInputMessage.ActionType, 2)
        Delete = typing___cast(TextInputMessage.ActionType, 3)
        ClearText = typing___cast(TextInputMessage.ActionType, 4)
    Unknown = typing___cast(TextInputMessage.ActionType, 0)
    Insert = typing___cast(TextInputMessage.ActionType, 1)
    Set = typing___cast(TextInputMessage.ActionType, 2)
    Delete = typing___cast(TextInputMessage.ActionType, 3)
    ClearText = typing___cast(TextInputMessage.ActionType, 4)

    timestamp = ... # type: float
    text = ... # type: typing___Text
    actionType = ... # type: TextInputMessage.ActionType

    def __init__(self,
        *,
        timestamp : typing___Optional[float] = None,
        text : typing___Optional[typing___Text] = None,
        actionType : typing___Optional[TextInputMessage.ActionType] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> TextInputMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    if sys.version_info >= (3,):
        def HasField(self, field_name: typing_extensions___Literal[u"actionType",u"text",u"timestamp"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"actionType",u"text",u"timestamp"]) -> None: ...
    else:
        def HasField(self, field_name: typing_extensions___Literal[u"actionType",b"actionType",u"text",b"text",u"timestamp",b"timestamp"]) -> bool: ...
        def ClearField(self, field_name: typing_extensions___Literal[u"actionType",b"actionType",u"text",b"text",u"timestamp",b"timestamp"]) -> None: ...

textInputMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import sys
from google.protobuf.descriptor import (
    Descriptor as google___protobuf___descriptor___Descriptor,
    FieldDescriptor as google___protobuf___descriptor___FieldDescriptor,
)

from google.protobuf.message import (
    Message as google___protobuf___message___Message,
)


class GetKeyboardSessionMessage(google___protobuf___message___Message):
    DESCRIPTOR: google___protobuf___descriptor___Descriptor = ...

    def __init__(self,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> GetKeyboardSessionMessage: ...
    def MergeFrom(self, other_msg: google___protobuf___message___Message) -> None: ...
    def CopyFrom(self, other_msg: google___protobuf___message___Message) -> None: ...

getKeyboardSessionMessage = ... # type: google___protobuf___descriptor___FieldDescriptor



"""Fake MRP Apple TV for tests."""

import asyncio
import logging
import struct
from datetime import datetime

from pyatv import const
from pyatv.mrp import (chacha20, messages, protobuf, variant)
from pyatv.mrp.protobuf import CommandInfo_pb2 as cmd
from pyatv.mrp.protobuf import SetStateMessage as ssm
from tests.airplay.fake_airplay_device import (
    FakeAirPlayDevice, AirPlayUseCases)
from tests.mrp.mrp_server_auth import MrpServerAuth

_LOGGER = logging.getLogger(__name__)

_KEY_LOOKUP = {
    # name: [usage_page, usage, button hold time (seconds)]
    'up': [1, 0x8C, 0],
    'down': [1, 0x8D, 0],
    'left': [1, 0x8B, 0],
    'right': [1, 0x8A, 0],
    'stop': [12, 0xB7, 0],
    'next': [12, 0xB5, 0],
    'previous': [12, 0xB6, 0],
    'select': [1, 0x89, 0],
    'menu': [1, 0x86, 0],
    'topmenu': [12, 0x60, 0],
    'home': [12, 0x40, 1],
    'suspend': [1, 0x82, 0],
    'wakeup': [1, 0x83, 0],
    'volume_up': [12, 0xE9, 0],
    'volume_down': [12, 0xEA, 0],
}

_COMMAND_LOOKUP = {
    cmd.Play: 'play',
    cmd.Pause: 'pause',
    cmd.Stop: 'stop',
    cmd.NextTrack: 'nextitem',
    cmd.PreviousTrack: 'previtem',
}

_REPEAT_LOOKUP = {
    const.RepeatState.Track: protobuf.CommandInfo.One,
    const.RepeatState.All: protobuf.CommandInfo.All,
}

_SHUFFLE_LOOKUP = {
    const.ShuffleState.Off: protobuf.CommandInfo.Off,
    const.ShuffleState.Albums: protobuf.CommandInfo.Albums,
    const.ShuffleState.Songs: protobuf.CommandInfo.Songs,
}

_COCOA_BASE = (datetime(1970, 1, 1) - datetime(2001, 1, 1)).total_seconds()


DEVICE_NAME = 'Fake MRP ATV'
PLAYER_IDENTIFIER = 'com.github.postlund.pyatv'


def _convert_key_press(use_page, usage):
    for name, codes in _KEY_LOOKUP.items():
        if codes[0] == use_page and codes[1] == usage:
            return name
    raise Exception(
        'unsupported key: use_page={0}, usage={1}'.format(
            use_page, usage))


def _fill_item(item, metadata):
    if metadata.identifier:
        item.identifier = metadata.identifier

    md = item.metadata
    if metadata.artist:
        md.trackArtistName = metadata.artist
    if metadata.album:
        md.albumName = metadata.album
    if metadata.title:
        md.title = metadata.title
    if metadata.genre:
        md.genre = metadata.genre
    if metadata.total_time:
        md.duration = metadata.total_time
    if metadata.position:
        md.elapsedTime = metadata.position
    if metadata.playback_rate:
        md.playbackRate = metadata.playback_rate
    if metadata.media_type:
        md.mediaType = metadata.media_type
    if metadata.artwork_mimetype:
        md.artworkAvailable = True
        md.artworkMIMEType = metadata.artwork_mimetype
        md.artworkIdentifier = metadata.artwork_identifier


def _set_state_message(metadata, identifier):
    # Most things are hardcoded here for simplicity. Will change that
    # as time goes by and more dynamic content is needed.
    set_state = messages.create(protobuf.SET_STATE_MESSAGE)
    inner = set_state.inner()
    inner.playbackState = metadata.playback_state
    inner.displayName = 'Fake Player'
    inner.playbackStateTimestamp = _COCOA_BASE

    if metadata.repeat and metadata.repeat != const.RepeatState.Off:
        cmd = inner.supportedCommands.supportedCommands.add()
        cmd.command = protobuf.CommandInfo_pb2.ChangeRepeatMode
        cmd.repeatMode = _REPEAT_LOOKUP[metadata.repeat]

    if metadata.shuffle:
        cmd = inner.supportedCommands.supportedCommands.add()
        cmd.command = protobuf.CommandInfo_pb2.ChangeShuffleMode
        cmd.shuffleMode = metadata.shuffle

    queue = inner.playbackQueue
    queue.location = 0
    _fill_item(queue.contentItems.add(), metadata)

    client = inner.playerPath.client
    client.processIdentifier = 123
    client.bundleIdentifier = identifier
    return set_state


class PlayingMetadata:

    def __init__(self, **kwargs):
        """Initialize a new PlayingMetadata."""
        self.identifier = kwargs.get('identifier')
        self.playback_state = kwargs.get('playback_state')
        self.title = kwargs.get('title')
        self.artist = kwargs.get('artist')
        self.album = kwargs.get('album')
        self.genre = kwargs.get('genre')
        self.total_time = kwargs.get('total_time')
        self.position = kwargs.get('position')
        self.repeat = kwargs.get('repeat')
        self.shuffle = _SHUFFLE_LOOKUP.get(kwargs.get('shuffle'))
        self.media_type = kwargs.get('media_type')
        self.playback_rate = kwargs.get('playback_rate')
        self.artwork = None
        self.artwork_mimetype = None


class FakeAppleTV(FakeAirPlayDevice, MrpServerAuth, asyncio.Protocol):
    """Implementation of a fake MRP Apple TV."""

    def __init__(self, testcase, loop):
        FakeAirPlayDevice.__init__(self, testcase)
        MrpServerAuth.__init__(self, self, DEVICE_NAME)
        self.loop = loop
        self.app.on_startup.append(self.start)
        self.outstanding_keypresses = set()  # Pressed but not released
        self.last_button_pressed = None
        self.connection_state = None
        self.states = {}
        self.active_player = None
        self.has_authenticated = False

        self.server = None
        self.buffer = b''
        self.chacha = None
        self.transport = None

    async def start(self, app):
        coro = self.loop.create_server(lambda: self, '127.0.0.1')
        self.server = await self.loop.create_task(coro)
        _LOGGER.info('Started MRP server at port %d', self.port)

    @property
    def port(self):
        return self.server.sockets[0].getsockname()[1]

    def connection_made(self, transport):
        self.transport = transport

    def enable_encryption(self, input_key, output_key):
        """Enable encryption with specified keys."""
        self.chacha = chacha20.Chacha20Cipher(
            input_key, output_key)
        self.has_authenticated = True

    def send(self, message):
        data = message.SerializeToString()
        if self.chacha:
            data = self.chacha.encrypt(data)

        length = variant.write_variant(len(data))
        self.transport.write(length + data)

    def update_state(self, identifier):
        state = self.states[identifier]
        self.send(_set_state_message(state, identifier))

    def set_player_state(self, identifier, state):
        self.states[identifier] = state
        self.update_state(identifier)

    def get_player_state(self, identifier):
        return self.states[identifier]

    def set_active_player(self, identifier):
        if identifier is not None and identifier not in self.states:
            raise Exception('invalid player: %s', identifier)

        self.active_player = identifier
        now_playing = messages.create(
            protobuf.SET_NOW_PLAYING_CLIENT_MESSAGE)
        client = now_playing.inner().client
        if identifier:
            client.bundleIdentifier = identifier
        self.send(now_playing)

    def item_update(self, metadata, identifier):
        msg = messages.create(protobuf.UPDATE_CONTENT_ITEM_MESSAGE)
        inner = msg.inner()

        _fill_item(inner.contentItems.add(), metadata)

        client = inner.playerPath.client
        client.processIdentifier = 123
        client.bundleIdentifier = identifier

        self.send(msg)

    def data_received(self, data):
        self.buffer += data

        while self.buffer:
            length, raw = variant.read_variant(self.buffer)
            if len(raw) < length:
                return

            data = raw[:length]
            self.buffer = raw[length:]
            if self.chacha:
                data = self.chacha.decrypt(data)

            parsed = protobuf.ProtocolMessage()
            parsed.ParseFromString(data)
            _LOGGER.info('Incoming message: %s', parsed)

            try:
                name = parsed.Type.Name(
                    parsed.type).lower().replace('_message', '')

                _LOGGER.debug('Received %s message', name)
                getattr(self, 'handle_' + name)(parsed, parsed.inner())
            except AttributeError:
                _LOGGER.exception('No message handler for ' + str(parsed))
            except Exception:
                _LOGGER.exception('Error while dispatching message')

    def handle_device_info(self, message, inner):
        resp = messages.device_information(DEVICE_NAME, '1234')
        resp.identifier = message.identifier
        self.send(resp)

    def handle_set_connection_state(self, message, inner):
        _LOGGER.debug('Changed connection state to %d', inner.state)
        self.connection_state = inner.state

    def handle_client_updates_config(self, message, inner):
        pass

    def handle_get_keyboard_session(self, message, inner):
        # This message has a lot more fields, but pyatv currently
        # not use them so ignore for now
        self.send(messages.create(
            protobuf.KEYBOARD_MESSAGE, identifier=message.identifier))

    def handle_send_hid_event(self, message, inner):
        # These corresponds to the bytes mapping to pressed key (see
        # send_hid_event in pyatv/mrp/messages.py)
        start = inner.hidEventData[43:49]
        use_page, usage, down_press = struct.unpack('>HHH', start)

        if down_press == 1:
            self.outstanding_keypresses.add((use_page, usage))
        elif down_press == 0:
            if (use_page, usage) in self.outstanding_keypresses:
                self.last_button_pressed = _convert_key_press(use_page, usage)
                self.outstanding_keypresses.remove((use_page, usage))
                _LOGGER.debug('Pressed button: %s', self.last_button_pressed)
            else:
                _LOGGER.error('Missing key down for %d,%d', use_page, usage)
        else:
            _LOGGER.error('Invalid key press state: %d', down_press)

    def handle_send_command(self, message, inner):
        button = _COMMAND_LOOKUP.get(inner.command)
        if button:
            self.last_button_pressed = button
            _LOGGER.debug('Pressed button: %s', self.last_button_pressed)
        elif inner.command == cmd.ChangeRepeatMode:
            state = self.get_player_state(self.active_player)
            repeatMode = inner.options.repeatMode
            if repeatMode == protobuf.CommandInfo.One:
                state.repeat = const.RepeatState.Track
            elif repeatMode == protobuf.CommandInfo.All:
                state.repeat = const.RepeatState.All
            else:
                state.repeat = const.RepeatState.Off
            self.update_state(self.active_player)
            _LOGGER.debug('Change repeat state to %s', state.repeat)
        elif inner.command == cmd.ChangeShuffleMode:
            state = self.get_player_state(self.active_player)
            state.shuffle = inner.options.shuffleMode
            self.update_state(self.active_player)
            _LOGGER.debug('Change shuffle state to %s', state.shuffle)
        elif inner.command == cmd.SeekToPlaybackPosition:
            state = self.get_player_state(self.active_player)
            state.position = inner.options.playbackPosition
            self.update_state(self.active_player)
            _LOGGER.debug('Seek to position: %d', state.position)
        else:
            _LOGGER.warning(
                'Unhandled button press: %s', message.inner().command)

    def handle_playback_queue_request(self, message, inner):
        setstate = messages.create(
            protobuf.SET_STATE_MESSAGE, identifier=message.identifier)
        queue = setstate.inner().playbackQueue
        queue.location = 0
        item = queue.contentItems.add()
        item.artworkData = self.states[self.active_player].artwork
        item.artworkDataWidth = 456
        item.artworkDataHeight = 789
        self.send(setstate)


class AppleTVUseCases(AirPlayUseCases):
    """Wrapper for altering behavior of a FakeMrpAppleTV instance."""

    def __init__(self, fake_apple_tv):
        """Initialize a new AppleTVUseCases."""
        self.device = fake_apple_tv

    def change_artwork(self, artwork, mimetype, identifier='artwork'):
        """Call this method to change artwork response."""
        metadata = self.device.get_player_state(PLAYER_IDENTIFIER)
        metadata.artwork = artwork
        metadata.artwork_mimetype = mimetype
        metadata.artwork_identifier = identifier
        self.device.update_state(PLAYER_IDENTIFIER)

    def change_metadata(self, **kwargs):
        """Change metadata for item via ContentItemUpdate."""
        metadata = self.device.get_player_state(PLAYER_IDENTIFIER)

        # Update saved metadata
        for key, value in kwargs.items():
            setattr(metadata, key, value)

        # Create a temporary metadata instance with requested parameters
        change = PlayingMetadata(**kwargs)
        self.device.item_update(change, PLAYER_IDENTIFIER)

    def nothing_playing(self):
        """Call this method to put device in idle state."""
        self.device.set_active_player(None)

    def example_video(self, **kwargs):
        """Play some example video."""
        self.video_playing(paused=True, title='dummy',
                           total_time=123, position=3, **kwargs)

    def video_playing(self, paused, title, total_time, position, **kwargs):
        """Call this method to change what is currently plaing to video."""
        metadata = PlayingMetadata(
            playback_state=ssm.Paused if paused else ssm.Playing,
            title=title, total_time=total_time,
            position=position, media_type=protobuf.ContentItemMetadata.Video,
            **kwargs)
        self.device.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.device.set_active_player(PLAYER_IDENTIFIER)

    def music_playing(self, paused, artist, album, title, genre,
                      total_time, position, **kwargs):
        """Call this method to change what is currently plaing to music."""
        metadata = PlayingMetadata(
            playback_state=ssm.Paused if paused else ssm.Playing,
            artist=artist, album=album, title=title, genre=genre,
            total_time=total_time, position=position,
            media_type=protobuf.ContentItemMetadata.Music,
            **kwargs)
        self.device.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.device.set_active_player(PLAYER_IDENTIFIER)

    def media_is_loading(self):
        """Call this method to put device in a loading state."""
        metadata = PlayingMetadata(playback_state=ssm.Interrupted)
        self.device.set_player_state(PLAYER_IDENTIFIER, metadata)
        self.device.set_active_player(PLAYER_IDENTIFIER)



"""Various helper methods."""

import asyncio
import pyatv


def auto_connect(handler, timeout=5, not_found=None):
    """Short method for connecting to a device.

    This is a convenience method that create an event loop, auto discovers
    devices, picks the first device found, connects to it and passes it to a
    user provided handler. An optional error handler can be provided that is
    called when no device was found. Very inflexible in many cases, but can be
    handys sometimes when trying things.

    Note: both handler and not_found must be coroutines
    """
    # A coroutine is used so we can connect to the device while being inside
    # the event loop
    async def _handle(loop):
        atvs = await pyatv.scan(loop, timeout=timeout)

        # Take the first device found
        if atvs:
            atv = await pyatv.connect(atvs[0], loop)
            try:
                await handler(atv)
            finally:
                await atv.logout()
        else:
            if not_found is not None:
                await not_found()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_handle(loop))



"""Various types of extraction and conversion functions."""

from pyatv import exceptions
from pyatv.const import (
    Protocol, MediaType, DeviceState, RepeatState, ShuffleState)


def media_kind(kind):
    """Convert iTunes media kind to API representation."""
    if kind in [1]:
        return MediaType.Unknown
    if kind in [3, 7, 11, 12, 13, 18, 32]:
        return MediaType.Video
    if kind in [2, 4, 10, 14, 17, 21, 36]:
        return MediaType.Music
    if kind in [8, 64]:
        return MediaType.TV

    raise exceptions.UnknownMediaKindError('Unknown media kind: ' + str(kind))


def media_type_str(mediatype):
    """Convert internal API media type to string."""
    if mediatype == MediaType.Unknown:
        return 'Unknown'
    if mediatype == MediaType.Video:
        return 'Video'
    if mediatype == MediaType.Music:
        return 'Music'
    if mediatype == MediaType.TV:
        return 'TV'
    return 'Unsupported'


def playstate(state):
    """Convert iTunes playstate to API representation."""
    # pylint: disable=too-many-return-statements
    if state == 0 or state is None:
        return DeviceState.Idle
    if state == 1:
        return DeviceState.Loading
    if state == 2:
        return DeviceState.Stopped
    if state == 3:
        return DeviceState.Paused
    if state == 4:
        return DeviceState.Playing
    if state in (5, 6):
        return DeviceState.Seeking

    raise exceptions.UnknownPlayStateError('Unknown playstate: ' + str(state))


# pylint: disable=too-many-return-statements
def playstate_str(state):
    """Convert internal API playstate to string."""
    if state == DeviceState.Idle:
        return 'Idle'
    if state == DeviceState.Loading:
        return 'Loading'
    if state == DeviceState.Stopped:
        return 'Stopped'
    if state == DeviceState.Paused:
        return 'Paused'
    if state == DeviceState.Playing:
        return 'Playing'
    if state == DeviceState.Seeking:
        return 'Seeking'

    return 'Unsupported'


def repeat_str(state):
    """Convert internal API repeat state to string."""
    if state == RepeatState.Off:
        return 'Off'
    if state == RepeatState.Track:
        return 'Track'
    if state == RepeatState.All:
        return 'All'
    return 'Unsupported'


def shuffle_str(state):
    """Convert internal API shuffle state to string."""
    if state == ShuffleState.Off:
        return 'Off'
    if state == ShuffleState.Albums:
        return 'Albums'
    if state == ShuffleState.Songs:
        return 'Songs'
    return 'Unsupported'


def ms_to_s(time):
    """Convert time in ms to seconds."""
    if time is None:
        return 0

    # Happens in some special cases, just return 0
    if time >= (2**32 - 1):
        return 0
    return round(time / 1000.0)


def protocol_str(protocol):
    """Convert internal API protocol to string."""
    if protocol == Protocol.MRP:
        return 'MRP'
    if protocol == Protocol.DMAP:
        return 'DMAP'
    if protocol == Protocol.AirPlay:
        return 'AirPlay'
    return 'Unknown'



"""Implementation of the DMAP protocol used by ATV 1, 2 and 3."""

import logging
import asyncio

from aiohttp.client_exceptions import ClientError

from pyatv import (exceptions, convert, net)
from pyatv.cache import Cache
from pyatv.const import Protocol, MediaType, RepeatState, ShuffleState
from pyatv.dmap import (parser, tags)
from pyatv.dmap.daap import DaapRequester
from pyatv.net import HttpSession
from pyatv.interface import (AppleTV, RemoteControl, Metadata,
                             Playing, PushUpdater, ArtworkInfo)

_LOGGER = logging.getLogger(__name__)

_PSU_CMD = 'ctrl-int/1/playstatusupdate?[AUTH]&revision-number={0}'
_ARTWORK_CMD = 'ctrl-int/1/nowplayingartwork?mw=1024&mh=576&[AUTH]'
_CTRL_PROMPT_CMD = 'ctrl-int/1/controlpromptentry?[AUTH]&prompt-id=0'


class BaseDmapAppleTV:
    """Common protocol logic used to interact with an Apple TV."""

    def __init__(self, requester):
        """Initialize a new Apple TV base implemenation."""
        self.daap = requester
        self.playstatus_revision = 0
        self.latest_hash = None

    async def playstatus(self, use_revision=False, timeout=None):
        """Request raw data about what is currently playing.

        If use_revision=True, this command will "block" until playstatus
        changes on the device.

        Must be logged in.
        """
        cmd_url = _PSU_CMD.format(
            self.playstatus_revision if use_revision else 0)
        resp = await self.daap.get(cmd_url, timeout=timeout)
        self.playstatus_revision = parser.first(resp, 'cmst', 'cmsr')
        playing = DmapPlaying(resp)
        self.latest_hash = playing.hash
        return playing

    async def artwork(self):
        """Return an image file (png) for what is currently playing.

        None is returned if no artwork is available. Must be logged in.
        """
        art = await self.daap.get(_ARTWORK_CMD, daap_data=False)
        return art if art != b'' else None

    def ctrl_int_cmd(self, cmd):
        """Perform a "ctrl-int" command."""
        cmd_url = 'ctrl-int/1/{}?[AUTH]&prompt-id=0'.format(cmd)
        return self.daap.post(cmd_url)

    def controlprompt_cmd(self, cmd):
        """Perform a "controlpromptentry" command."""
        data = tags.string_tag('cmbe', cmd) + tags.uint8_tag('cmcc', 0)
        return self.daap.post(_CTRL_PROMPT_CMD, data=data)

    def controlprompt_data(self, data):
        """Perform a "controlpromptentry" command."""
        return self.daap.post(_CTRL_PROMPT_CMD, data=data)

    def set_property(self, prop, value):
        """Change value of a DAAP property, e.g. volume or media position."""
        cmd_url = 'ctrl-int/1/setproperty?{}={}&[AUTH]'.format(
            prop, value)
        return self.daap.post(cmd_url)


# pylint: disable=too-many-public-methods
class DmapRemoteControl(RemoteControl):
    """Implementation of API for controlling an Apple TV."""

    def __init__(self, apple_tv):
        """Initialize remote control instance."""
        super().__init__()
        self.apple_tv = apple_tv

    async def up(self):
        """Press key up."""
        await self._send_commands(
            self._move('Down', 0, 20, 275),
            self._move('Move', 1, 20, 270),
            self._move('Move', 2, 20, 265),
            self._move('Move', 3, 20, 260),
            self._move('Move', 4, 20, 255),
            self._move('Move', 5, 20, 250),
            self._move('Up', 6, 20, 250))

    async def down(self):
        """Press key down."""
        await self._send_commands(
            self._move('Down', 0, 20, 250),
            self._move('Move', 1, 20, 255),
            self._move('Move', 2, 20, 260),
            self._move('Move', 3, 20, 265),
            self._move('Move', 4, 20, 270),
            self._move('Move', 5, 20, 275),
            self._move('Up', 6, 20, 275))

    async def left(self):
        """Press key left."""
        await self._send_commands(
            self._move('Down', 0, 75, 100),
            self._move('Move', 1, 70, 100),
            self._move('Move', 3, 65, 100),
            self._move('Move', 4, 60, 100),
            self._move('Move', 5, 55, 100),
            self._move('Move', 6, 50, 100),
            self._move('Up', 7, 50, 100))

    async def right(self):
        """Press key right."""
        await self._send_commands(
            self._move('Down', 0, 50, 100),
            self._move('Move', 1, 55, 100),
            self._move('Move', 3, 60, 100),
            self._move('Move', 4, 65, 100),
            self._move('Move', 5, 70, 100),
            self._move('Move', 6, 75, 100),
            self._move('Up', 7, 75, 100))

    @staticmethod
    def _move(direction, time, point1, point2):
        data = 'touch{0}&time={1}&point={2},{3}'.format(
            direction, time, point1, point2)
        return tags.uint8_tag('cmcc', 0x30) + tags.string_tag('cmbe', data)

    async def _send_commands(self, *cmds):
        for cmd in cmds:
            await self.apple_tv.controlprompt_data(cmd)

    def play(self):
        """Press key play."""
        return self.apple_tv.ctrl_int_cmd('play')

    def pause(self):
        """Press key pause."""
        return self.apple_tv.ctrl_int_cmd('pause')

    def stop(self):
        """Press key stop."""
        return self.apple_tv.ctrl_int_cmd('stop')

    def next(self):
        """Press key next."""
        return self.apple_tv.ctrl_int_cmd('nextitem')

    def previous(self):
        """Press key previous."""
        return self.apple_tv.ctrl_int_cmd('previtem')

    def select(self):
        """Press key select."""
        return self.apple_tv.controlprompt_cmd('select')

    def menu(self):
        """Press key menu."""
        return self.apple_tv.controlprompt_cmd('menu')

    def top_menu(self):
        """Press key topmenu."""
        return self.apple_tv.controlprompt_cmd('topmenu')

    def volume_up(self):
        """Press key volume up."""
        # DMAP support unknown
        raise exceptions.NotSupportedError()

    def volume_down(self):
        """Press key volume down."""
        # DMAP support unknown
        raise exceptions.NotSupportedError()

    def home(self):
        """Press key home."""
        # DMAP support unknown
        raise exceptions.NotSupportedError()

    def home_hold(self):
        """Hold key home."""
        # DMAP support unknown
        raise exceptions.NotSupportedError()

    def suspend(self):
        """Suspend the device."""
        # Not supported by DMAP
        raise exceptions.NotSupportedError()

    def wakeup(self):
        """Wake up the device."""
        raise exceptions.NotSupportedError()

    def set_position(self, pos):
        """Seek in the current playing media."""
        time_in_ms = int(pos)*1000
        return self.apple_tv.set_property('dacp.playingtime', time_in_ms)

    def set_shuffle(self, shuffle_state):
        """Change shuffle mode to on or off."""
        state = 0 if shuffle_state == ShuffleState.Off else 1
        return self.apple_tv.set_property('dacp.shufflestate', state)

    def set_repeat(self, repeat_state):
        """Change repeat mode."""
        return self.apple_tv.set_property(
            'dacp.repeatstate', repeat_state.value)


class DmapPlaying(Playing):
    """Implementation of API for retrieving what is playing."""

    def __init__(self, playstatus):
        """Initialize playing instance."""
        super().__init__()
        self.playstatus = playstatus

    @property
    def media_type(self):
        """Type of media is currently playing, e.g. video, music."""
        state = parser.first(self.playstatus, 'cmst', 'caps')
        if not state:
            return MediaType.Unknown

        mediakind = parser.first(self.playstatus, 'cmst', 'cmmk')
        if mediakind is not None:
            return convert.media_kind(mediakind)

        # Fallback: if artist or album exists we assume music (not present
        # for video)
        if self.artist or self.album:
            return MediaType.Music

        return MediaType.Video

    @property
    def device_state(self):
        """Device state, e.g. playing or paused."""
        state = parser.first(self.playstatus, 'cmst', 'caps')
        return convert.playstate(state)

    @property
    def title(self):
        """Title of the current media, e.g. movie or song name."""
        return parser.first(self.playstatus, 'cmst', 'cann')

    @property
    def artist(self):
        """Arist of the currently playing song."""
        return parser.first(self.playstatus, 'cmst', 'cana')

    @property
    def album(self):
        """Album of the currently playing song."""
        return parser.first(self.playstatus, 'cmst', 'canl')

    @property
    def genre(self):
        """Genre of the currently playing song."""
        return parser.first(self.playstatus, 'cmst', 'cang')

    @property
    def total_time(self):
        """Total play time in seconds."""
        return self._get_time_in_seconds('cast')

    @property
    def position(self):
        """Position in the playing media (seconds)."""
        return self.total_time - self._get_time_in_seconds('cant')

    @property
    def shuffle(self):
        """If shuffle is enabled or not."""
        state = parser.first(self.playstatus, 'cmst', 'cash')
        if state is None or state == 0:
            return ShuffleState.Off

        # DMAP does not support the "albums" state and will always report
        # "songs" if shuffle is active
        return ShuffleState.Songs

    @property
    def repeat(self):
        """Repeat mode."""
        state = parser.first(self.playstatus, 'cmst', 'carp')
        if state is None:
            return RepeatState.Off
        return RepeatState(state)

    def _get_time_in_seconds(self, tag):
        time = parser.first(self.playstatus, 'cmst', tag)
        return convert.ms_to_s(time)


class DmapMetadata(Metadata):
    """Implementation of API for retrieving metadata from an Apple TV."""

    def __init__(self, identifier, apple_tv):
        """Initialize metadata instance."""
        super().__init__(identifier)
        self.apple_tv = apple_tv
        self.artwork_cache = Cache(limit=4)

    async def artwork(self):
        """Return artwork for what is currently playing (or None)."""
        # Having to fetch "playing" here is not ideal, but an identifier is
        # needed and we cannot trust any previous identifiers. So we have to do
        # this until a better solution comes along.
        playing = await self.playing()
        identifier = playing.hash
        if identifier in self.artwork_cache:
            _LOGGER.debug('Retrieved artwork %s from cache', identifier)
            return self.artwork_cache.get(identifier)

        artwork = await self._fetch_artwork()
        if artwork:
            self.artwork_cache.put(identifier, artwork)
            return artwork

        return None

    async def _fetch_artwork(self):
        _LOGGER.debug('Fetching artwork')
        data = await self.apple_tv.artwork()
        if data:
            return ArtworkInfo(data, 'image/png')
        return None

    @property
    def artwork_id(self):
        """Return a unique identifier for current artwork."""
        return self.apple_tv.latest_hash

    async def playing(self):
        """Return current device state."""
        return await self.apple_tv.playstatus()


class DmapPushUpdater(PushUpdater):
    """Implementation of API for handling push update from an Apple TV."""

    def __init__(self, loop, apple_tv, listener):
        """Initialize a new DmapPushUpdater instance."""
        super().__init__()
        self._loop = loop
        self._atv = apple_tv
        self._listener = listener
        self._future = None
        self._initial_delay = 0

    @property
    def active(self):
        """Return if push updater has been started."""
        return self._future is not None

    def start(self, initial_delay=0):
        """Wait for push updates from device.

        Will throw NoAsyncListenerError if no listener has been set.
        """
        if self.listener is None:
            raise exceptions.NoAsyncListenerError()
        if self.active:
            return

        # Always start with 0 to trigger an immediate response for the
        # first request
        self._atv.playstatus_revision = 0

        # Delay before restarting after an error
        self._initial_delay = initial_delay

        self._future = asyncio.ensure_future(self._poller(), loop=self._loop)

    def stop(self):
        """No longer forward updates to listener."""
        if self._future is not None:
            self._future.cancel()
            self._future = None

    async def _poller(self):
        first_call = True

        while True:
            # Sleep some time before waiting for updates
            if not first_call and self._initial_delay > 0:
                _LOGGER.debug('Initial delay set to %d', self._initial_delay)
                await asyncio.sleep(self._initial_delay, loop=self._loop)
                first_call = False

            try:
                _LOGGER.debug('Waiting for playstatus updates')
                playstatus = await self._atv.playstatus(
                    use_revision=True, timeout=0)

                self._loop.call_soon(
                    self.listener.playstatus_update, self, playstatus)
            except asyncio.CancelledError:
                break

            except ClientError as ex:
                _LOGGER.exception('A communication error happened')
                if self._listener and self._listener.listener:
                    self._loop.call_soon(
                        self._listener.listener.connection_lost, ex)

                break

            # It is not pretty to disable pylint here, but we must catch _all_
            # exceptions to keep the API.
            except Exception as ex:  # pylint: disable=broad-except
                _LOGGER.debug('Playstatus error occurred: %s', ex)
                self._loop.call_soon(self.listener.playstatus_error, self, ex)

        self._future = None


class DmapAppleTV(AppleTV):
    """Implementation of API support for Apple TV."""

    # This is a container class so it's OK with many attributes
    # pylint: disable=too-many-instance-attributes
    def __init__(self, loop, session, config, airplay):
        """Initialize a new Apple TV."""
        super().__init__()
        self._session = session

        self._dmap_service = config.get_service(Protocol.DMAP)
        daap_http = HttpSession(
            session,
            'http://{0}:{1}/'.format(config.address, self._dmap_service.port))
        self._requester = DaapRequester(
            daap_http, self._dmap_service.credentials)

        self._apple_tv = BaseDmapAppleTV(self._requester)
        self._dmap_remote = DmapRemoteControl(self._apple_tv)
        self._dmap_metadata = DmapMetadata(config.identifier, self._apple_tv)
        self._dmap_push_updater = DmapPushUpdater(loop, self._apple_tv, self)
        self._airplay = airplay

    def connect(self):
        """Initiate connection to device.

        No need to call it yourself, it's done automatically.
        """
        return self._requester.login()

    async def close(self):
        """Close connection and release allocated resources."""
        if net.is_custom_session(self._session):
            await self._session.close()
        if self.listener:
            self.listener.connection_closed()

    @property
    def service(self):
        """Return service used to connect to the Apple TV.."""
        return self._dmap_service

    @property
    def remote_control(self):
        """Return API for controlling the Apple TV."""
        return self._dmap_remote

    @property
    def metadata(self):
        """Return API for retrieving metadata from Apple TV."""
        return self._dmap_metadata

    @property
    def push_updater(self):
        """Return API for handling push update from the Apple TV."""
        return self._dmap_push_updater

    @property
    def stream(self):
        """Return API for streaming media."""
        return self._airplay



"""Configuration when connecting to an Apple TV."""
from pyatv import (convert, exceptions)
from pyatv.const import Protocol


class AppleTV:
    """Representation of an Apple TV configuration.

    An instance of this class represents a single device. A device can have
    several services, depending on the protocols it supports, e.g. DMAP or
    AirPlay.
    """

    def __init__(self, address, name):
        """Initialize a new AppleTV."""
        self.address = address
        self.name = name
        self._services = {}

    @property
    def identifier(self):
        """Return one of the identifiers associated with this device."""
        for prot in [Protocol.MRP, Protocol.DMAP, Protocol.AirPlay]:
            service = self.get_service(prot)
            if service:
                return service.identifier
        return None

    @property
    def all_identifiers(self):
        """Return all unique identifiers for this device."""
        services = self.services
        return [x.identifier for x in services if x.identifier is not None]

    def add_service(self, service):
        """Add a new service.

        If the service already exists, it will be merged.
        """
        existing = self._services.get(service.protocol)
        if existing is not None:
            existing.merge(service)
        else:
            self._services[service.protocol] = service

    def get_service(self, protocol):
        """Look up a service based on protocol.

        If a service with the specified protocol is not available, None is
        returned.
        """
        return self._services.get(protocol)

    @property
    def services(self):
        """Return all supported services."""
        services = set(self._services.keys())
        return [self.get_service(x) for x in services]

    def main_service(self, protocol=None):
        """Return suggested service used to establish connection."""
        protocols = [protocol] if protocol is not None else \
            [Protocol.MRP, Protocol.DMAP]

        for prot in protocols:
            service = self.get_service(prot)
            if service is not None:
                return service

        raise exceptions.NoServiceError(
            'no service to connect to')

    def set_credentials(self, protocol, credentials):
        """Set credentials for a protocol if it exists."""
        service = self.get_service(protocol)
        if service:
            service.credentials = credentials
            return True
        return False

    def __eq__(self, other):
        """Compare instance with another instance."""
        if isinstance(other, self.__class__):
            return self.identifier == other.identifier
        return False

    def __str__(self):
        """Return a string representation of this object."""
        services = [' - {0}'.format(s) for s in self._services.values()]
        identifiers = [' - {0}'.format(x) for x in self.all_identifiers]
        return '       Name: {0}\n' \
               '    Address: {1}\n' \
               'Identifiers:\n' \
               '{2}\n' \
               'Services:\n' \
               '{3}'.format(self.name, self.address,
                            '\n'.join(identifiers), '\n'.join(services))


# pylint: disable=too-few-public-methods
class BaseService:
    """Base class for protocol services."""

    def __init__(self, identifier, protocol, port):
        """Initialize a new BaseService."""
        self.__identifier = identifier
        self.protocol = protocol
        self.port = port
        self.credentials = None

    @property
    def identifier(self):
        """Return unique identifier associated with this service."""
        return self.__identifier

    def merge(self, other):
        """Merge with other service of same type."""
        self.credentials = other.credentials or self.credentials

    def __str__(self):
        """Return a string representation of this object."""
        return 'Protocol: {0}, Port: {1}, Credentials: {2}'.format(
            convert.protocol_str(self.protocol), self.port,
            self.credentials)


# pylint: disable=too-few-public-methods
class DmapService(BaseService):
    """Representation of a DMAP service."""

    def __init__(self, identifier, credentials, port=None):
        """Initialize a new DmapService."""
        super().__init__(identifier, Protocol.DMAP, port or 3689)
        self.credentials = credentials


# pylint: disable=too-few-public-methods
class MrpService(BaseService):
    """Representation of a MediaRemote Protocol service."""

    def __init__(self, identifier, port, credentials=None):
        """Initialize a new MrpService."""
        super().__init__(identifier, Protocol.MRP, port)
        self.credentials = credentials


# pylint: disable=too-few-public-methods
class AirPlayService(BaseService):
    """Representation of an AirPlay service."""

    def __init__(self, identifier, port=7000, credentials=None):
        """Initialize a new AirPlayService."""
        super().__init__(identifier, Protocol.AirPlay, port)
        self.credentials = credentials



"""Methods used to GET/POST data from/to an Apple TV."""

import re
import logging

from copy import copy


from pyatv import exceptions
from pyatv.dmap import parser
from .tag_definitions import lookup_tag

_LOGGER = logging.getLogger(__name__)

_DMAP_HEADERS = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip',
    'Client-DAAP-Version': '3.13',
    'Client-ATV-Sharing-Version': '1.2',
    'Client-iTunes-Sharing-Version': '3.15',
    'User-Agent': 'Remote/1021',
    'Viewer-Only-Client': '1',
}


DEFAULT_TIMEOUT = 10.0  # Seconds


class DaapRequester:
    """Helper class that makes it easy to perform DAAP requests.

    It will automatically do login and other necesarry book-keeping.
    """

    def __init__(self, http, login_id):
        """Initialize a new DaapRequester."""
        self.http = http
        self._login_id = login_id
        self._session_id = 0

    async def login(self):
        """Login to Apple TV using specified login id."""
        # Do not use session.get_data(...) in login as that would end up in
        # an infinte loop.
        def _login_request():
            return self.http.get_data(
                self._mkurl('login?[AUTH]&hasFP=1',
                            session=False, login_id=True),
                headers=_DMAP_HEADERS)

        resp = await self._do(_login_request, is_login=True)
        self._session_id = parser.first(resp, 'mlog', 'mlid')

        _LOGGER.info('Logged in and got session id %s', self._session_id)
        return self._session_id

    async def get(self, cmd, daap_data=True, timeout=None, **args):
        """Perform a DAAP GET command."""
        def _get_request():
            return self.http.get_data(
                self._mkurl(cmd, *args),
                headers=_DMAP_HEADERS,
                timeout=timeout)

        await self._assure_logged_in()
        return await self._do(_get_request, is_daap=daap_data)

    async def post(self, cmd, data=None, timeout=None, **args):
        """Perform DAAP POST command with optional data."""
        def _post_request():
            headers = copy(_DMAP_HEADERS)
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
            return self.http.post_data(
                self._mkurl(cmd, *args),
                data=data,
                headers=headers,
                timeout=timeout)

        await self._assure_logged_in()
        return await self._do(_post_request)

    async def _do(self, action, retry=True, is_login=False, is_daap=True):
        resp, status = await action()
        if is_daap:
            resp = parser.parse(resp, lookup_tag)

        self._log_response(str(action.__name__) + ': %s', resp, is_daap)
        if 200 <= status < 300:
            return resp

        if not is_login:
            # If a request fails, try to login again before retrying
            _LOGGER.info('implicitly logged out, logging in again')
            await self.login()

        # Retry once if we got a bad response, otherwise bail out
        if retry:
            return (await self._do(
                action, False, is_login=is_login, is_daap=is_daap))

        raise exceptions.AuthenticationError(
            'failed to login: ' + str(status))

    def _mkurl(self, cmd, *args, session=True, login_id=False):
        url = '{}'.format(cmd.format(*args))
        parameters = []
        if login_id:
            if re.match(r'0x[0-9a-fA-F]{16}', self._login_id):
                parameters.append('pairing-guid={}'.format(self._login_id))
            else:
                parameters.append('hsgid={}'.format(self._login_id))
        if session:
            parameters.insert(0, 'session-id={}'.format(self._session_id))
        return url.replace('[AUTH]', '&'.join(parameters))

    async def _assure_logged_in(self):
        if self._session_id != 0:
            _LOGGER.debug('Already logged in, re-using seasion id %d',
                          self._session_id)
        else:
            await self.login()

    @staticmethod
    def _log_response(text, data, is_daap):
        if _LOGGER.isEnabledFor(logging.INFO):
            formatted = data
            if is_daap:
                formatted = parser.pprint(data, lookup_tag)
            _LOGGER.debug(text, formatted)
